"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import contextlib
import io
import re
import os
import sys
import tempfile
import shutil
import logging
import functools
import warnings
from collections import OrderedDict
import numpy
from . import Backend
from .. import __homepage__
from ..loss import Loss
from ..utils import can_import, EnvironmentalVariable, redirect_stderr, \
	idx, DisableLogging
from ..providers import BatchProvider

logger = logging.getLogger(__name__)

###############################################################################
class KerasBackend(Backend):
	""" A Keras backend.

		# Dependencies

		- keras
		- theano OR tensorflow
		- h5py
	"""

	###########################################################################
	@classmethod
	def is_supported(cls):
		""" Returns True if this backend can be used.
		"""
		return can_import('keras') and (
			can_import('theano') or can_import('tensorflow')
		)

	###########################################################################
	def __init__(self, backend=None, optimizer=None, theano_flags=None,
		*args, **kwargs):
		""" Creates a new Keras backend.

			As per the base class documentation, we should do all necessary
			Keras-related initialization here, including checking obvious
			things like "Is Keras installed?" or "Is a backend installed?"

			# Arguments

			backend: str or None (default: None). The Keras backend to use
				(either "theano" or "tensorflow"). None uses the system
				default.
			optimizer: None or False (default: None). If False, Theano is told
				to disable optimizations. If None, Theano is not told to do
				anything special with optimization. This is supplied as a
				workaround to installing a BLAS library when training on the
				CPU. This is ignored for the TensorFlow backend.
		"""

		super().__init__(*args, **kwargs)

		if backend is not None:
			logger.info('The %s backend for Keras has been requested.',
				backend)

			if 'keras' in sys.modules:
				import keras.backend as K		# pylint: disable=import-error
				if K.backend() != backend:
					logger.warning('Keras was already imported by the time '
						'the Kur backend was instantiated. Kur was asked to '
						'use Keras %s backend, but Keras is already using %s. '
						'We cannot change the Keras backend at this point, so '
						'we will try to work with the currently loaded '
						'backend. In the future, try to let Kur manage '
						'importing Keras.', backend, K.backend())

			deps = {
				'theano' : ['theano'],
				'tensorflow' : ['tensorflow']
			}[backend]
			for dep in deps:
				if can_import(dep):
					continue
				if backend == 'tensorflow':
					logger.error('Your Kurfile is trying to use TensorFlow.')
					logger.error('However, we cannot find TensorFlow '
						'installed.')
					logger.error('At least it is easy to install!')
					logger.error('To install TensorFlow for CPU: pip install '
						'tensorflow')
					logger.error('To install TensorFlow for GPU: pip install '
						'tensorflow-gpu')
					logger.error('See our troubleshooting page for more '
						'information: %s', os.path.join(__homepage__,
						'troubleshooting.html'))
					raise ValueError('Need to install TensorFlow for this '
						'Kurfile to work.')
				else:
					logger.warning('The Keras backend was asked to use the %s '
						'backend, but %s does not appear to be installed. You '
						'will likely get an error about this soon.',
						backend, dep)

		else:
			logger.info('No particular backend for Keras has been requested.')
			if can_import('theano') and can_import('tensorflow'):
				logger.debug('Using the system-default Keras backend.')
			elif can_import('theano'):
				backend = 'theano'
				logger.debug('Only the Theano backend for Keras is installed, '
					'so we will try to use it.')
			elif can_import('tensorflow'):
				backend = 'tensorflow'
				logger.debug('Only the TensorFlow backend for Keras is '
					'installed, so we will try to use it.')
			else:
				logger.warning('No supported Keras backend seems to be '
					'installed. You will probably get an error about this '
					'shortly.')

		# Make sure Keras is loaded.
		# Now, Keras always prints out a "Using {Theano|TensorFlow} backend."
		# statement that is frankly unbecoming. So we'll just gobble it up here.
		x = io.StringIO()
		with redirect_stderr(x):

			env = {
				'KERAS_BACKEND' : backend,
				'THEANO_FLAGS' : os.environ.get('THEANO_FLAGS')
			}

			def replace_theano_flag(key, value):
				""" Updates the Theano flag variable.
				"""
				if env['THEANO_FLAGS']:
					parts = [i for i in env['THEANO_FLAGS'].split(',') \
						if not i.startswith('{}='.format(key))]
				else:
					parts = []
				parts.append('{}={}'.format(key, value))
				env['THEANO_FLAGS'] = ','.join(parts)

			if optimizer is False:
				logger.debug('Disabling the Theano optimizer.')
				replace_theano_flag('optimizer', 'None')

			if theano_flags is not None:
				for k, v in theano_flags.items():
					logger.debug('Setting Theano flag %s = %s', k, v)
					replace_theano_flag(k, v)

			replace_theano_flag('force_device', 'true')
			if not self.devices:
				replace_theano_flag('device', 'cpu')
				env['CUDA_VISIBLE_DEVICES'] = '100'
				logger.info('Requesting CPU')
			else:
				replace_theano_flag('device', 'gpu')
				env['CUDA_VISIBLE_DEVICES'] = ','.join(
					str(x) for x in self.devices)
				logger.info('Requesting GPUs: %s', self.devices)

			# Supress the deluge of TensorFlow messages that we aren't
			# interested in.
			env['TF_CPP_MIN_LOG_LEVEL'] = '1'

			logger.debug('Overriding environmental variables: %s', env)
			EnvironmentalVariable(**env).push()

			import keras	# pylint: disable=import-error,unused-variable
			import keras.backend as K		# pylint: disable=import-error
			logger.info('Keras is loaded. The backend is: %s',
				K.backend())
			self.toolchain = K.backend()

		# And now we can set the dimension ordering.
		keras.backend.set_image_dim_ordering('tf')

		# The Keras `Wrapper` class accesses `Layer`'s `regularizers`
		# property (see `Wrapper.build()`), which triggers Keras' own
		# deprecation warning. Let's suppress this for now so that we don't
		# confuse our users.
		logging.getLogger('py.warnings').addFilter(
			type('theano_filter', (), {
				'filter' : lambda record: not (
					record.module == 'topology' and
					record.levelname == 'WARNING' and
					record.funcName == 'regularizers'
				)
			})
		)

		if self.parallel > 1 and self.get_toolchain() == 'theano':
			logger.warning('Multiple GPUs were requested, but are not '
				'supported with Keras\' Theano backend. Try the PyTorch '
				'backend or Keras\' TensorFlow backend instead. Falling back '
				'to a single device.')
			self.devices = self.devices[:1]

	###########################################################################
	def get_toolchain(self):
		""" Returns a string describing the Keras backend being used, either
			'theano' or 'tensorflow'.
		"""
		return self.toolchain

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the backend class.

			This is used by the base class's factory method.
		"""
		return 'keras'

	###########################################################################
	def connect(self, inputs, target, data):
		""" Use the Keras functional API to connect to layers

			# Notes:

			- You will need input placeholders in place before doing this,
			  otherwise Keras's shape-checking will fail.
		"""
		if self.keras_version() == 1:
			if not isinstance(inputs, list):
				inputs = [inputs]
		else:
			if isinstance(inputs, (list, tuple)):
				if len(inputs) == 1:
					inputs = inputs[0]

		pool_2d = None
		if self.get_toolchain() == 'theano':
			import theano						# pylint: disable=import-error
			if theano.__version__ < '0.9':
				# We need to patch Theano
				from theano.tensor.signal import pool # pylint: disable=import-error
				original_pool = pool.pool_2d
				def pool_2d(input, ws=None, ignore_border=None, stride=None,
						pad=(0, 0), mode='max', ds=None, st=None,
						padding=None):
					return original_pool(
						input=input,
						ds=ds if ds is not None else ws,
						ignore_border=ignore_border,
						st=st if st is not None else stride,
						padding=padding if padding is not None else pad,
						mode=mode
					)

		with warnings.catch_warnings():
			warnings.filterwarnings(
				'ignore',
				message='.*tensor.nnet.abstract_conv.conv2d.*',
				module='.*theano_backend.*'
			)
			if pool_2d is None:
				return target(inputs)
			else:
				from unittest.mock import patch
				with patch('theano.tensor.signal.pool.pool_2d', pool_2d):
					return target(inputs)

	###########################################################################
	@staticmethod
	def keras_version():
		""" Retrieves the Keras major version.
		"""
		from keras import __version__			# pylint: disable=import-error
		return int(__version__.split('.')[0])

	###########################################################################
	@staticmethod
	def make_model(inputs, outputs):
		""" Compiles a Keras model in a version-agnostic way.
		"""
		import keras.models as M				# pylint: disable=import-error
		if KerasBackend.keras_version() == 1:
			return M.Model(input=inputs, output=outputs)
		return M.Model(inputs=inputs, outputs=outputs)

	###########################################################################
	def save(self, model, filename):
		""" Saves the model weights to the given filename.
		"""
		keras_model = self.make_model(
			inputs=[node.value for node in model.inputs.values()],
			outputs=[node.value for node in model.outputs.values()]
		)

		self._save_keras(keras_model, filename)

	###########################################################################
	def _save_keras(self, keras_model, filename):
		""" Saves a native Keras model.
		"""
		path = os.path.expanduser(os.path.expandvars(filename))
		if os.path.exists(path):
			if not os.path.isdir(path):
				raise ValueError('Target weight exists, but it is not a '
					'directory. Kur expected a directory that it can work '
					'with. Please move or delete the existing path: {}'
					.format(path))

			for dirpath, _, filenames in os.walk(path):
				for this_file in filenames:
					if this_file.endswith('.kur'):
						os.unlink(os.path.join(dirpath, this_file))
		else:
			os.makedirs(path, exist_ok=True)

		layers = keras_model.flattened_layers \
			if hasattr(keras_model, 'flattened_layers') else keras_model.layers
		for layer in layers:
			layer_name = layer.name

			symbolic_weights = layer.weights
			weight_names, weight_values = \
				self._get_weight_names_and_values_from_symbolic(
					symbolic_weights
				)

			for name, val in zip(weight_names, weight_values):
				name = name.replace('/', '_')
				target = os.path.join(
					path,
					'{}+{}.kur'.format(layer_name, name)
				)
				idx.save(target, val)

	###########################################################################
	def _get_weight_names_and_values_from_symbolic(self, symbolic_weights):
		import keras.backend as K				# pylint: disable=import-error
		weight_values = K.batch_get_value(symbolic_weights)
		weight_names = [
			(
				str(w.name) if hasattr(w, 'name') and w.name \
					else 'param_{}'.format(i)
			)
			for i, (w, val) in enumerate(
				zip(symbolic_weights, weight_values)
			)
		]
		return weight_names, weight_values

	###########################################################################
	def restore(self, model, filename):
		""" Load the model weights from the given filename.
		"""
		keras_model = self.make_model(
			inputs=[node.value for node in model.inputs.values()],
			outputs=[node.value for node in model.outputs.values()]
		)

		try:
			self._restore_keras(keras_model, filename)
		except:
			logger.exception('Failed to load previously-saved Keras model. '
				'Are you accidentally loading pre-existing weights from an '
				'incompatible model? Make sure that any weights on disk are '
				'actually associated with the model you are trying to load '
				'them into.')
			raise

	###########################################################################
	def _restore_keras(self, keras_model, filename):
		""" Loads a native Keras model.
		"""
		import keras.backend as K				# pylint: disable=import-error

		path = os.path.expanduser(os.path.expandvars(filename))
		if os.path.exists(path):
			if not os.path.isdir(path):
				raise ValueError('Target weight exists, but it is not a '
					'directory. Kur expected a directory that it can work '
					'with. Please move or delete the existing path: {}'
					.format(path))
		else:
			raise ValueError('Target weight directory does not exist: {}'
				.format(path))

		layers = keras_model.flattened_layers \
			if hasattr(keras_model, 'flattened_layers') else keras_model.layers

		# Get a map from "layer name" to "layer instance" in the current model.
		index = {}
		for layer in layers:
			if layer.name:
				index.setdefault(layer.name, []).append(layer)

		# Enumerate all of the saved tensors, organized like this:
		# tensors = {
		#	'layer_1_name' : {
		#		'weight_1_name'  : '/path/to/file',
		#		...,
		#	},
		#	...
		# }
		tensors = self.enumerate_saved_tensors(path)

		# We want to put (symbolic_weights, weight_values) tuples in this.
		weight_value_tuples = []

		# Loop over the available weights.
		for layer_name, weights in tensors.items():

			# Load the weights.
			# This maps weight names to numpy arrays.
			weights = {k : idx.load(v) for k, v in weights.items()}

			# Now assign all of the weights to their corresponding symbolic
			# weights. Loop over all layers which use this name.
			for layer in index.get(layer_name, []):

				# Get the symbolic weights.
				symbolic_weights = layer.weights

				# Get the associated names (so we know what order to assign the
				# weights in.
				weight_names, _ = \
					self._get_weight_names_and_values_from_symbolic(
						symbolic_weights
					)

				available = set(weights.keys())
				needed = set(name.replace('/', '_') for name in weight_names)
				if available ^ needed:
					logger.error('Weight discrepancy in the weights we are '
						'supposed to load.')
					logger.error('These weights are on-disk, but not '
						'requested: %s', ', '.join(available - needed))
					logger.error('These weights were requested, but not '
						'available: %s', ', '.join(needed - available))
					raise ValueError('Layer "{}" expected {} weights, but we '
						'found {} on disk.'.format(layer_name,
						len(needed), len(available)))

				for i, name in enumerate(weight_names):
					name = name.replace('/', '_')
					weight_value_tuples.append((symbolic_weights[i], weights[name]))

		# Assign all the weights in one batch (for efficiency).
		K.batch_set_value(weight_value_tuples)

	###########################################################################
	def enumerate_saved_tensors(self, path):
		""" Enumerates saved tensors (weights).
		"""
		result = {}

		regex = re.compile(r'^(?P<layer>.*?)\+(?P<weight>.*?)\.kur$')
		for dirpath, dirnames, filenames in os.walk(path): # pylint: disable=unused-variable
			for filename in filenames:
				match = regex.match(filename)
				if match is None:
					continue
				filename = os.path.join(dirpath, filename)
				layer, weight = match.groups()
				if layer not in result:
					result[layer] = {}
				if weight in result[layer]:
					logger.warning('Tensor weights have already been loaded '
						'for layer="%s", tensor="%s" from file: %s. We will '
						'skip this new file we just found: %s.', layer, weight,
						result[layer][weight], filename)
					continue
				result[layer][weight] = filename

		return result

	###########################################################################
	@staticmethod
	def find_compiled_layer_by_name(model, layer_name):
		""" Returns the Keras tensor associated with a given name.

			# Arguments

			model: Model instance. The Kur model. It must be compiled.
			layer_name: str. The name of the layer to find, or one of its
				aliases.

			# Return value

			The Keras tensor
		"""
		if model.compiled is None or 'raw' not in model.compiled:
			raise ValueError('The model must be compiled first.')
		if layer_name in model.output_aliases:
			target = model.output_aliases[layer_name]
		elif layer_name in model.input_aliases:
			target = model.input_aliases[layer_name]
		else:
			raise ValueError('Failed to find a layer named "{}"'
				.format(layer_name))
		for keras_layer, kur_layer in \
			zip(model.compiled['raw'].outputs, model.outputs):
			if kur_layer == target:
				return keras_layer
		raise ValueError('Did not find expected layer. This is a bug.')

	###########################################################################
	def process_loss(self, model, loss):
		""" Process the loss functions.

			# Arguments

			model: The Kur model. It must be compiled.
			loss: Loss instance, list/tuple of Loss instances, or a dictionary
				of model layer names mapped to Loss instances.
		"""
		import keras.backend as K				# pylint: disable=import-error

		if loss is None:
			num_outputs = len(model.outputs)
			logger.error('You are trying to construct a training/validation'
				'/testing model, but you haven\'t specified any loss '
				'functions. Your model has %d outputs: %s. You need to '
				'specify %d loss functions, one for each output.',
				num_outputs, ', '.join(model.outputs), num_outputs)
			raise ValueError('No loss functions were specified, but are '
				'required for training, testing, and validation.')

		if isinstance(loss, Loss):
			loss = [loss]

		if len(loss) != len(model.outputs):
			raise ValueError('Model has {} outputs, but only {} loss '
				'functions were specified.'
				.format(len(model.outputs), len(loss)))

		if isinstance(loss, (list, tuple)):
			loss = dict(zip(model.outputs, loss))

		if not isinstance(loss, (dict, OrderedDict)):
			raise ValueError('Loss functions given to "compile" should be '
				'a list/tuple, a dictionary, or a single Loss instance. '
				'Instead we received this: {} (type={})'
				.format(loss, type(loss)))

		loss_inputs = OrderedDict()
		loss_outputs = OrderedDict()
		for target, this_loss in loss.items():
			ins, out = this_loss.get_loss(
				model,
				target,
				self.find_compiled_layer_by_name(model, target)
			)
			# FIXME: Re-using a network output in different loss functions will
			# probably break, since each loss function is creating its own
			# placeholder inputs, but then we are throwing some away using
			# 'update'.
			loss_inputs.update(ins)
			loss_outputs[target] = K.mean(out)
			logger.debug('Adding additional inputs: %s',
				', '.join(x[0] for x in ins))

		total_loss = functools.reduce(
			lambda x, y: x + y,
			loss_outputs.values()
		)
		return loss_inputs, loss_outputs, total_loss

	###########################################################################
	def compile(self, model, loss=None, optimizer=None, blocking=True,
		assemble_only=False):
		""" Returns the Keras model instance.
		"""
		if model.compiled is None:
			model.compiled = {}

		if 'raw' not in model.compiled:
			logger.debug('Instantiating a Keras model.')
			compiled = self.make_model(
				inputs=[node.value for node in model.inputs.values()],
				outputs=[node.value for node in model.outputs.values()]
			)

			if logger.isEnabledFor(logging.DEBUG):
				x = io.StringIO()
				with contextlib.redirect_stdout(x):
					compiled.summary()
				for line in x.getvalue().split('\n'):
					logger.debug(line)

			if self.parallel > 1:
				from ..utils.parallelism import make_parallel
				compiled = make_parallel(compiled, self.parallel)

			model.compiled['raw'] = compiled

		else:
			logger.debug('Reusing an existing model.')
			compiled = model.compiled['raw']

		import keras.backend as K				# pylint: disable=import-error
		if loss is None and optimizer is None:
			logger.debug('Assembling an evaluation function from the model.')

			loss_inputs = loss_outputs = {}
			if not assemble_only:
				func = K.function(
					compiled.inputs + \
						[K.learning_phase()],
					compiled.outputs
				)
			key = 'evaluate'

		elif optimizer is None:
			logger.debug('Assembling a testing function from the model.')

			loss_inputs, loss_outputs, _ = \
				self.process_loss(model, loss)

			if not assemble_only:
				func = K.function(
					compiled.inputs + \
						list(loss_inputs.values()) + \
						[K.learning_phase()],
					compiled.outputs + \
						list(loss_outputs.values())
				)
			key = 'test'

		else:
			logger.debug('Assembling a training function from the model.')

			# Loss inputs: additional inputs needed by the loss function.
			# Loss outputs: output of the loss function
			loss_inputs, loss_outputs, total_loss = \
				self.process_loss(model, loss)

			updates = optimizer.get_optimizer(self)(
				compiled.trainable_weights, total_loss
			)

			if not assemble_only:
				func = K.function(
					compiled.inputs + \
						list(loss_inputs.values()) + \
						[K.learning_phase()],
					compiled.outputs + \
						list(loss_outputs.values()),
					updates=updates
				)
			key = 'train'

		logger.debug('Additional inputs for log functions: %s',
			', '.join(loss_inputs.keys()))

		input_names = compiled.input_names + \
			list(loss_inputs.keys())
		output_names = compiled.output_names + \
			list(loss_outputs.keys())

		input_shapes = [
			layer._keras_shape
			for layer in compiled.inputs
		] + [
			layer._keras_shape
			for layer in loss_inputs.values()
		]

		logger.debug('Expected input shapes: %s',
			', '.join('{}={}'.format(k, v) for k, v in \
				zip(input_names, input_shapes)
			))

		if assemble_only:
			func = None

		result = {
			'func' : func,
			'names' : {
				'input' : input_names,
				'output' : output_names
			},
			'shapes' : {
				'input' : input_shapes
			}
		}

		if logger.isEnabledFor(logging.DEBUG):
			logger.debug('Compiled model: %s', result)

		if not assemble_only:
			model.compiled[key] = result
			if blocking:
				self.wait_for_compile(model, key)

		return result

	###########################################################################
	def wait_for_compile(self, model, key):
		""" Waits for the model to finish compiling.
		"""
		if model.provider is None:
			logger.warning('No data provider available, so we cannot reliably '
				'wait for compiling to finish.')
			return

		with DisableLogging():
			provider = BatchProvider(
				sources=dict(zip(model.provider.keys, model.provider.sources)),
				batch_size=2*max(1, self.parallel),
				num_batches=1,
				randomize=False
			)
		model.supplement_provider(provider)

		weight_path = None
		tempdir = tempfile.mkdtemp()
		try:
			weight_path = os.path.join(tempdir, 'weights')
			self._save_keras(model.compiled['raw'], weight_path)

			logger.info('Waiting for model to finish compiling...')
			for batch in provider:
				self.run_batch(model, batch, key, False)

		finally:
			if weight_path and os.path.isdir(weight_path):
				try:
					self._restore_keras(model.compiled['raw'], weight_path)
				except:
					logger.error('We were waiting for the model to finish '
						'compiling, but failed to restore the model weights. '
						'The weights may be in a bad state.')
					raise

			shutil.rmtree(tempdir, ignore_errors=True)

	###########################################################################
	def run_batch(self, model, batch, key, is_train):
		if model.compiled is None or key not in model.compiled:
			raise ValueError('A model has not been compiled to: {}'
				.format(key))

		compiled = model.compiled[key]
		raw = model.compiled['raw']

		assert isinstance(is_train, bool)

		def coerce_shape(data, shape, name):
			if data.ndim < len(shape):
				return numpy.expand_dims(data, -1)
			else:
				return data

		inputs = [
			coerce_shape(
				batch[model.get_data_name_by_layer_name(batch, name)],
				shape, name
			)
			for shape, name in zip(
				compiled['shapes']['input'],
				compiled['names']['input']
			)
		] + [is_train]

		outputs = compiled['func'](inputs)
		num_outputs = len(raw.outputs)
		metrics = {
			k : v for k, v in zip(
				compiled['names']['output'][num_outputs:],
				outputs[num_outputs:]
			)
		}
		predictions = {name : data for name, data in zip(model.outputs, outputs[:num_outputs])}
		return predictions, metrics

	###########################################################################
	def train(self, model, data):
		""" Fits the given model on a batch of data.
		"""
		return self.run_batch(model, data, 'train', True)

	###########################################################################
	def test(self, model, data):
		""" Calculates the model loss on a batch of data.
		"""
		return self.run_batch(model, data, 'test', False)

	###########################################################################
	def evaluate(self, model, data):
		""" Evaluates the model on a batch of data.
		"""
		return self.run_batch(model, data, 'evaluate', False)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy
from . import ChunkSource

###############################################################################
class RepeatSource(ChunkSource):
	""" A data source which always returns the same value.
	"""

	###########################################################################
	def __init__(self, value=None, *args, **kwargs):
		""" Create a new repeat data source.

			# Arguments

			value: object. The value to return for each sample.
			chunk_size: int (default: None). The maximum number of samples to
				return at each iteration, or None to choose some default
				number.
		"""
		super().__init__(*args, **kwargs)

		self.data = numpy.array([value for _ in range(self.chunk_size)])

	###########################################################################
	def __len__(self):
		""" Returns the number of samples this source provides.
		"""
		return 0

	###########################################################################
	def shape(self):
		""" Returns the shape of the individual data samples.
		"""
		return self.data.shape[1:]

	###########################################################################
	def can_shuffle(self):
		""" This source can be shuffled.
		"""
		return True

	###########################################################################
	def shuffle(self, indices):
		""" Applies a permutation to the data.
		"""
		pass

	###########################################################################
	def __iter__(self):
		""" Returns the next chunk of data.
		"""

		while True:
			yield self.data

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Transpose(Layer):				# pylint: disable=too-few-public-methods
	""" A transpose layer changes the ordering of the dimensions.

		# Usage

		```
		transpose: [0, 1]
		```

		or

		```
		transpose:
		  axes: [0, 1, 2]
		  include_batch: yes
		```
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new transpose layer.
		"""
		super().__init__(*args, **kwargs)
		self.axes = None
		self.include_batch = False

	###########################################################################
	def _parse(self, engine):
		""" Parses out the transpose layer.
		"""
		if isinstance(self.args, (list, tuple)):
			self.axes = self.args
			self.include_batch = False
		elif isinstance(self.args, dict):
			if 'axes' in self.args:
				self.axes = engine.evaluate(self.args['axes'])
			else:
				raise ParsingError('Missing required key "axes" for transpose '
					'layer.')
			if 'include_batch' in self.args:
				self.include_batch = engine.evaluate(
					self.args['include_batch'])
			else:
				self.include_batch = False

		if not isinstance(self.axes, (list, tuple)):
			raise ParsingError('Value for "axes" in transpose layer must be a '
				'list or tuple. Received: {}'.format(self.axes))

		if not isinstance(self.include_batch, bool):
			raise ParsingError('Value for "batch" in transpose layer must be a '
				'boolean. Received: {}'.format(self.include_batch))

		for axis in self.axes:
			if not isinstance(axis, int):
				raise ParsingError('All axes in transpose layer must be '
					'integers. Received: {}'.format(axis))
			if axis < 0:
				raise ParsingError('All axes in transpose layer must be >= 0.')
		if set(self.axes) != set(range(len(self.axes))):
			raise ParsingError('Invalid axes transposition: {}'
				.format(self.axes))

	###########################################################################
	def _build(self, model):
		""" Instantiates the layer with the given backend.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			import keras.layers as L			# pylint: disable=import-error
			import keras.backend as K			# pylint: disable=import-error

			if self.include_batch:
				dims = self.axes
			else:
				dims = (0, ) + tuple(x+1 for x in self.axes)

			def transpose_shape(input_shape):
				""" Computes the expanded shape.
				"""
				return tuple(input_shape[i] for i in dims)

			yield L.Lambda(
				lambda x: K.permute_dimensions(x, dims),
				transpose_shape,
				name=self.name
			)

		elif backend.get_name() == 'pytorch':

			if self.include_batch:
				dims = self.axes
			else:
				dims = (0, ) + tuple(x+1 for x in self.axes)

			def connect(inputs):
				""" Connects the layer.
				"""
				assert len(inputs) == 1
				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : model.data.add_operation(
						lambda x: x.permute(*dims)
					)(inputs[0]['layer'])
				}

			yield connect

		else:
			raise ValueError('Unknown or unsupported backend: {}'
				.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Transpose layers only take a single input.')
		input_shape = input_shapes[0]

		if self.include_batch:
			return tuple(input_shape[i-1] for i in self.axes if i > 0)
		return tuple(input_shape[i] for i in self.axes)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging
import shutil
import math
import time
import traceback
import numpy
import tqdm
from ..utils import get_any_value, CriticalSection, parallelize
from ..loggers import PersistentLogger
from .hooks import TrainingHook

logger = logging.getLogger(__name__)

###############################################################################
class RetryException(Exception):
	""" Exception class for retrying an operation on a new batch of data.
	"""
	pass

###############################################################################
class Executor:
	""" Class for using models.
	"""

	MAX_RETRIES = 3
	DEFAULT_RETRY_ENABLED = True
	NAN_IS_FATAL = True

	###########################################################################
	def __init__(self, model, loss=None, optimizer=None, auto_retry=None):
		""" Creates a new executor.

			# Arguments

			model: Model instance. The model to train.
			loss: Loss instance. The loss function to use in training/testing.
			optimizer: Optimizer instance. The optimizer to use in training.
		"""
		self.model = model
		self.loss = loss
		self.optimizer = optimizer

		if auto_retry is None:
			auto_retry = self.DEFAULT_RETRY_ENABLED
		self.auto_retry = auto_retry

	###########################################################################
	def compile(self, target=None, recompile=False, with_provider=None,
		**kwargs):
		""" Compiles a model.

			This generates a backend-specific representation of the model,
			suitable for training.

			# Arguments

			recompile: bool (default: False). If the model has already been
				compiled, it is not compiled again unless this flag is True.
			with_provider: Provider instance or None (default: None). If you
				want to merge the model's auxiliary data sources into your
				provider, you can specify the Provider instance here.

			# Return value

			None
		"""

		if target is None:
			if self.loss is None and self.optimizer is None:
				target = 'evaluate'
			elif self.optimizer is None:
				target = 'test'
			else:
				target = 'train'

		if not recompile:
			if self.model.compiled is not None \
				and target in self.model.compiled:
				return

		if not self.model.is_built():
			logger.warning('This model has never been built before. We are '
				'going to try to build it now. But the model should always be '
				'built with Model.build() before trying to compile it, just '
				'to ensure that everything has been parsed as you expect.')
			if with_provider is not None:
				self.model.register_provider(with_provider)
			self.model.build()

		logger.debug('Recompiling the model.')
		self.model.backend.compile(
			model=self.model,
			loss=self.loss if target != 'evaluate' else None,
			optimizer=None if target != 'train' else self.optimizer,
			blocking=True,
			**kwargs
		)

		if with_provider is not None:
			self.model.supplement_provider(with_provider)

	###########################################################################
	def test(self, provider, validating=False, hooks=None, step=False):
		""" Tests/validates the model on some data.

			# Arguments

			provider: Provider instance. The data provider which serves the
				data to be evaluated on.
			validating: bool (default: False). If False, the console output
				refers to this process as "testing"; otherwise, it is referred
				to as "validating."

			# Return value

			The average loss across the validation set.
		"""

		self.compile('test', with_provider=provider)

		if validating:
			desc = ('Validating', 'Validation')
		else:
			desc = ('Testing', 'Test')

		# Create progress bar
		test_loss = None
		n_entries = 0
		first_batch = None
		test_func = self.retry(
			self.model.backend.test,
			self.auto_retry
		)
		with tqdm.tqdm(
					total=len(provider),
					unit='samples',
					desc='{}, loss=N/A'.format(desc[0])
				) as pbar:

			# Present each batch to the network.
			for num_batches, batch in parallelize(enumerate(provider)):
				if step:
					self.do_step('Test', num_batches, batch)

				try:
					prediction, batch_loss = test_func(
						model=self.model,
						data=batch
					)
				except RetryException:
					continue

				if step and logger.isEnabledFor(logging.DEBUG):
					print(prediction)

				if first_batch is None:
					first_batch = (prediction, batch)

				batch_size = len(get_any_value(batch))

				#batch_loss = loss if isinstance(loss, float) \
				#	else sum(loss.values())

				new_entries = n_entries + batch_size

				if test_loss is None:
					test_loss = batch_loss
				else:
					test_loss = {
						k : v * (n_entries / new_entries) + \
							batch_loss[k] * (batch_size / new_entries)
						for k, v in test_loss.items()
					}
				#avg_loss = avg_loss * (n_entries / new_entries) + \
				#	batch_loss * (batch_size / new_entries)
				n_entries = new_entries

				# Update the progress bar
				pbar.set_description('{}, loss={:.3f}'.format(
					desc[0],
					sum(test_loss.values())
				))
				pbar.update(batch_size)

		if not n_entries:
			logger.warning('No data provided to validation/testing system.')
			return None

		logger.info('%s loss: %.3f', desc[1], sum(test_loss.values()))

		if hooks and first_batch is not None:
			prediction, batch = first_batch
			prev = first_batch
			for hook in hooks:
				new_prev = hook.apply(prev, first_batch, self.model)
				prev = (new_prev, prev[1]) \
					if not isinstance(new_prev, tuple) else new_prev

		return test_loss

	###########################################################################
	def train(self, *args, last_weights=None, log=None, training_hooks=None,
		**kwargs):
		""" Trains the model on some data.

			This is the public entry point for training. It wraps the business
			logic so that it can handle error conditions.
		"""

		reason = 'unknown'
		try:
			result = self.wrapped_train(
				*args,
				log=log,
				training_hooks=training_hooks,
				**kwargs
			)
		except (KeyboardInterrupt, Exception) as exc:
			logger.exception('Exception raised during training.')
			reason = traceback.format_exception_only(type(exc), exc)[0].strip()
			raise
		else:
			reason = 'success'
			return result
		finally:
			if last_weights is not None:
				logger.info('Saving most recent weights: %s', last_weights)
				with CriticalSection():
					self.model.save(last_weights)
			if log is not None:
				log.flush()

			if training_hooks:
				for hook in training_hooks:
					hook.notify(
						TrainingHook.TRAINING_END,
						log=log,
						info={'Reason' : reason}
					)

	###########################################################################
	def wrapped_train(self, provider, *, validation=None, epochs=None,
		log=None, best_train=None, best_valid=None, training_hooks=None,
		validation_hooks=None, checkpoint=None, step=False):
		""" Trains the model on some data.

			# Arguments

			provider: Provider instance. The data provider which serves the
				data to be trained on.
			validation: Provider instance or None (default: None). The data
				provider which serves validation data.
			epochs: int or None (default: None). The number of epochs to train
				for, or None to train forever.
			log: Log instance or None (default: None). The logger to save
				training statistics with.

			# Return value

			None
		"""

		#######################################################################
		# Process checkpoint requirements
		if isinstance(checkpoint, dict):
			if 'path' not in checkpoint:
				checkpoint['path'] = 'checkpoint'

			found = False
			for k in ('epochs', 'batches', 'samples'):
				if k in checkpoint:
					if not isinstance(checkpoint[k], int):
						raise ValueError('Expected "{}" key in "checkpoint" '
							'to be an integer. Received: {}'.format(k,
							checkpoint[k]))
					found = True

			if not found:
				checkpoint['epochs'] = 1

		elif isinstance(checkpoint, str):
			checkpoint = {
				'path' : checkpoint,
				'epochs' : 1
			}
		elif checkpoint is not None:
			raise ValueError('Unknown format for "checkpoint". Expected a '
				'single file or a dictionary. Instead we received: {}'
				.format(checkpoint))

		#######################################################################
		# Parse logs
		if log is None:
			logger.info('No log specified, so no historical loss information '
				'is available.')
			best_train_loss = best_valid_loss = None
		elif not isinstance(log, PersistentLogger):
			logger.info('Log type is non-persistent, so no historical loss '
				'information is available.')
			best_train_loss = best_valid_loss = None
		else:
			best_train_loss = log.get_best_training_loss()
			if best_train_loss is not None:
				logger.info('Best historical training loss: %.3f',
					best_train_loss)
			else:
				logger.info('No historical training loss available from logs.')

			best_valid_loss = log.get_best_validation_loss()
			if best_valid_loss is not None:
				logger.info('Best historical validation loss: %.3f',
					best_valid_loss)
			else:
				logger.info(
					'No historical validation loss available from logs.')

		#######################################################################
		# Parse desired number of epochs
		completed_epochs = log.get_number_of_epochs() if log else 0
		if not completed_epochs:
			logger.info('No previous epochs.')
		else:
			logger.info('Restarting from epoch %d.', completed_epochs+1)

		valid_modes = ('total', 'additional')
		default_mode = 'additional'
		mode = default_mode
		if isinstance(epochs, dict):
			mode = epochs.get('mode', default_mode)
			if mode not in valid_modes:
				raise ValueError('If "mode" in "epochs" must be one of: {}. '
					'Instead, we received: {}.'.format(', '.join(valid_modes),
					mode))
			if mode == 'total' and log is None:
				logger.warning('The epoch specification has "mode" set to '
					'"%s". This mode requires a log to be used correctly. Kur '
					'will proceed as if "mode" were "%s".', mode, default_mode)
				mode = default_mode
			epochs = epochs.get('number')
			if epochs in ('inf', 'all', 'infinite', 'infinity'):
				epochs = None
		elif not isinstance(epochs, (int, type(None))):
			raise ValueError('Expected "epochs" to be a dictionary or '
				'integer. Instead, we received: {}.'.format(epochs))
		logger.debug('Epoch handling mode: %s', mode)

		if epochs is not None:
			if mode == 'additional':
				epochs += completed_epochs

		#######################################################################
		# Local variables

		# The name of the most recently saved weight file. If the weights
		# change, this should be reset to None. Otherwise, saving weights can
		# be as simple as copying the previously saved file.
		saved_recent = None

		session = {
			'epochs' : 0,
			'batches' : 0,
			'samples' : 0,
			'minutes' : time.perf_counter() / 60
		}
		last_checkpoint = session.copy()

		epoch = completed_epochs - 1
		train_func = self.retry(
			self.model.backend.train,
			self.auto_retry
		)

		#######################################################################
		def run_validation(num_batches=None):
			""" Executes a validation run.
			"""
			if validation is None:
				return None

			nonlocal best_valid_loss

			# Continue with a validation run.
			try:
				if num_batches is not None and \
						hasattr(validation, 'num_batches'):
					previous_num_batches = validation.num_batches
					validation.num_batches = num_batches

				validation_loss = self.test(
					provider=validation,
					validating=True,
					hooks=validation_hooks
				)
			finally:
				if num_batches is not None and \
						hasattr(validation, 'num_batches'):
					validation.num_batches = previous_num_batches

			if validation_loss is None:
				return None

			cur_validation_loss = sum(validation_loss.values())
			if best_valid is not None:
				if best_valid_loss is None or \
						cur_validation_loss < best_valid_loss:
					logger.info(
						'Saving best historical validation weights: %s',
						best_valid
					)
					best_valid_loss = cur_validation_loss
					save_or_copy_weights(best_valid)

			if log is not None:
				log.log_validation(validation_loss, 'loss')

			return validation_loss

		#######################################################################
		def save_or_copy_weights(target):
			""" Saves the current model weights.
			"""
			nonlocal saved_recent

			if saved_recent is None:
				logger.debug('Saving weights to: %s', target)
				with CriticalSection():
					self.model.save(target)
				saved_recent = target
			elif not os.path.exists(saved_recent):
				logger.warning('Recently saved weight file seems to have '
					'vanished: %s', saved_recent)
				saved_recent = None
				save_or_copy_weights(target)
			elif os.path.exists(target) and \
					os.path.samefile(target, saved_recent):
				logger.debug('Recent weight file seems the same as the '
					'soon-to-be-saved file. Skipping: %s', target)
			else:
				logger.debug('Copying weights from: %s', saved_recent)
				with CriticalSection():
					shutil.rmtree(target, ignore_errors=True)
					shutil.copytree(saved_recent, target)

		#######################################################################
		def run_posttrain(n_entries, train_loss):
			""" Calculates training loss and saves if necessary.

				Read-only non-locals:
					n_entries, train_loss, best_train, log
				Read-write non-locals:
					best_train_loss
			"""
			nonlocal best_train_loss
			if not n_entries:
				logger.warning('No data provided to training loop.')
				return None

			cur_train_loss = sum(train_loss.values())
			logger.info('Training loss: %.3f', cur_train_loss)

			if best_train is not None:
				if best_train_loss is None or \
					cur_train_loss < best_train_loss:

					logger.info('Saving best historical training weights: '
						'%s', best_train)
					best_train_loss = cur_train_loss
					save_or_copy_weights(best_train)

			if log is not None:
				log.log_training(train_loss, 'loss')

			return cur_train_loss

		#######################################################################
		def run_training_hooks(cur_train_loss, validation_loss, status):
			""" Executes the training hooks, if necessary.

				Read-only non-locals:
					training_hooks, epoch, epochs, validation_loss
			"""
			if not training_hooks:
				return
			info = {
				'epoch' : epoch+1,
				'total_epochs' : epochs,
				'Training loss' : cur_train_loss
			}
			if validation is not None:
				info['Validation loss'] = validation_loss
			for hook in training_hooks:
				hook.notify(
					status,
					log=log,
					info=info
				)

		#######################################################################
		def run_checkpoint(*triggers, allow_validation=True):
			""" Runs the checkpoint triggers, if necessary.
			"""
			nonlocal last_checkpoint

			if checkpoint is None:
				return

			for k in triggers:
				if k not in checkpoint:
					continue
				if session[k] - last_checkpoint[k] >= checkpoint[k]:
					# We need a checkpoint

					# Save the file if necessary.
					if checkpoint['path']:
						logger.info('Making checkpoint backup: %s',
							checkpoint['path'])
						save_or_copy_weights(checkpoint['path'])

					# Validate if necessary.
					if checkpoint.get('validation', False) \
							and allow_validation:
						if isinstance(checkpoint['validation'], bool):
							num_batches = None
						else:
							num_batches = checkpoint['validation']
						val_loss = run_validation(num_batches)
						run_training_hooks(None, val_loss,
							TrainingHook.VALIDATION_END)

					last_checkpoint = session.copy()
					break

		#######################################################################
		# Prepare to train

		self.compile('train', with_provider=provider)
		provider.source_shapes()

		if training_hooks:
			for hook in training_hooks:
				hook.notify(
					TrainingHook.TRAINING_START,
					log=log
				)

		#######################################################################
		# Main training loop.
		while True:
			epoch += 1
			if epochs is not None and epoch >= epochs:
				print('Completed {} epochs.'.format(epochs))
				break

			print()

			###################################################################
			# START: Train one epoch

			# Create progress bar
			train_loss = None
			n_entries = 0
			with tqdm.tqdm(
						total=len(provider),
						unit='samples',
						desc='Epoch {}/{}, loss=N/A'
							.format(epoch+1, epochs or 'inf')
					) as pbar:

				# Present each batch to the network.
				for num_batches, batch in parallelize(enumerate(provider)):

					# The loss averaged over this batch.
					logger.debug('Training on batch...')
					if step:
						self.do_step(
							'Train, Epoch {}'.format(session['epochs']+1),
							num_batches, batch)

					try:
						prediction, batch_loss = train_func(
							model=self.model, data=batch)
					except RetryException:
						continue

					if step and logger.isEnabledFor(logging.DEBUG):
						print(prediction)

					# We just modified the weights. Invalidate the name of the
					# last weight file.
					saved_recent = None

					logger.debug('Finished training on batch.')

					# How many entries we just processed.
					batch_size = len(get_any_value(batch))

					if log is not None:
						log.log_batch(batch_size, batch_loss, 'loss')

					# Update our session statistics.
					session['batches'] += 1
					session['samples'] += batch_size
					session['minutes'] = time.perf_counter() / 60

					# Checkpoint if necessary
					run_checkpoint('samples', 'batches', 'minutes',
						allow_validation=True)

					# How many entries we've processed this epoch.
					new_entries = n_entries + batch_size

					# Average the per-batch loss across training.
					# This will give us our average "training loss".
					if train_loss is None:
						train_loss = batch_loss
					else:
						train_loss = {
							k : v * (n_entries / new_entries) + \
								batch_loss[k] * (batch_size / new_entries)
							for k, v in train_loss.items()
						}

					n_entries = new_entries

					# Update the progress bar with the current loss.
					# Note that `batch_loss` is, in some sense, just the
					# instantaneous training loss. `train_loss` is the average
					# loss across the entire training set so far.
					pbar.set_description('Epoch {}/{}, loss={:.3f}'.format(
						epoch+1, epochs or 'inf', sum(train_loss.values())
					))
					pbar.update(batch_size)

					for k, v in batch_loss.items():
						if math.isnan(v):
							logger.error('Received NaN loss value for '
								'model output "%s". Make sure that your '
								'inputs are all normalized and that the '
								'learning rate is not too high. Sometimes '
								'different algorithms/implementations '
								'work better than others, so you can try '
								'switching optimizers or backend.', k)
							if self.NAN_IS_FATAL:
								raise ValueError('Model loss is NaN.')

			# END: Train one epoch
			###################################################################

			# Update our session statistics.
			session['epochs'] += 1

			# Checkpoint if necessary
			run_checkpoint('epochs', allow_validation=False)

			# Check to see what our current training loss is.
			cur_train_loss = run_posttrain(n_entries, train_loss)

			# Validate
			validation_loss = run_validation()

			# Execute training hooks.
			run_training_hooks(
				cur_train_loss,
				validation_loss,
				status=TrainingHook.EPOCH_END
			)

	###########################################################################
	def evaluate(self, provider, callback=None, step=False):
		""" Evaluates the model on some data.

			# Arguments

			provider: Provider instance. The data provider which serves the
				data to be evaluated.
			callback: function or None. If not None, the callback is called
				after each evaluation batch and is passed two parameters:
				`predicted` and `truth`, where `predicted` is the model output
				and `truth` is the ground truth data (if provided by
				`provider`; otherwise, `truth` is set to `None`).

			# Return value

			If `callback` is None, then this returns a tuple `(predicted,
			truth)`, where `predicted` is a dictionary whose keys are the names
			of the output nodes of the model, and whose respective values are
			arrays of predictions (one row per input sample). If the provider
			provides ground truth information, then `truth` has a similar
			structure to `predicted`; if ground truth information is not
			available, then `truth` is None.

			Otherwise, if `callback` is not None, this returns None.
		"""

		self.compile('evaluate', with_provider=provider)

		result = None
		truth = None
		has_truth = None
		total = len(provider)
		n_entries = 0

		#######################################################################
		def store_batch_unknown(batch, evaluated, batch_size):
			""" Saves the batch if we do not know how many entries to expect.
			"""
			nonlocal truth, result

			# We don't know how many entries there will be.
			if result is None:
				# This is our first batch.
				result = {k : [] for k in self.model.outputs}
			for k, v in evaluated.items():
				result[k].extend(v)

			if has_truth:
				if truth is None:
					truth = {k : [] for k in self.model.outputs}
				for k in truth:
					truth[k].extend(batch[k])

		#######################################################################
		def store_batch_known(batch, evaluated, batch_size):
			""" Saves the batch if we know how many entries to expect.
			"""
			nonlocal truth, result

			# We know how many entries there will be.
			if result is None:
				# This is our first batch.
				result = {k : [None]*total for k in evaluated}
			for k, v in evaluated.items():
				result[k][n_entries:(n_entries+batch_size)] = v[:]

			if has_truth:
				if truth is None:
					truth = {k : [None]*total for k in evaluated}
				for k in truth:
					truth[k][n_entries:(n_entries+batch_size)] = batch[k][:]

		store_batch = store_batch_unknown if total is None \
			else store_batch_known

		eval_func = self.retry(
			self.model.backend.evaluate,
			self.auto_retry
		)

		with tqdm.tqdm(
					total=total,
					unit='samples',
					desc='Evaluating'
				) as pbar:

			for num_batches, batch in parallelize(enumerate(provider)):

				if step:
					self.do_step('Evaluate', num_batches, batch)

				try:
					evaluated, _ = eval_func(model=self.model, data=batch)
				except RetryException:
					continue

				if step and logger.isEnabledFor(logging.DEBUG):
					print(evaluated)

				batch_size = len(get_any_value(batch))

				# Check to see if we have truth data available.
				if has_truth is None:
					has_truth = all(k in batch for k in self.model.outputs)

				if callback is None:
					# There is no callback. We need to hang on to everything.
					store_batch(batch, evaluated, batch_size)
				else:
					callback(evaluated, truth)

				n_entries += batch_size
				pbar.update(batch_size)

		if callback is not None:
			return

		for data in (result, truth):
			if data is not None:
				for k, v in data.items():
					data[k] = numpy.array(v)

		if truth is not None:
			for k, v in truth.items():
				result[k] = numpy.reshape(result[k], v.shape)

		return result, truth

	###########################################################################
	def do_step(self, what, num_batches, batch):
		""" Wait for user input before running a single batch of data.
		"""
		print('{}, Batch {}:'.format(what, num_batches+1))
		if logger.isEnabledFor(logging.DEBUG):
			for k, v in batch.items():
				print('{} {}: {}'.format(
					k,
					v.shape if hasattr(v, 'shape') else \
						'(list, {} entries)'.format(len(v)),
					v
				))
		input('Press ENTER to continue...')

	###########################################################################
	def retry(self, func, enabled=True):
		""" Creates a wrapper that implements some retry semantics.
		"""

		def try_func(*args, **kwargs):
			""" Wraps a function with some retry logic.
			"""
			try:
				result = func(*args, **kwargs)

			# Catch Exception so that we don't catch KeyboardInterrupt.
			except Exception:
				if not try_func.enabled:
					raise

				try_func.counter += 1
				if try_func.counter > Executor.MAX_RETRIES:
					logger.exception(
						'Failed to execute on batch. No more retries.')
					raise
				logger.exception('Failed to execute on batch. Tolerating up '
					'to %d more consecutive failures.',
					Executor.MAX_RETRIES - try_func.counter)
				raise RetryException
			else:
				try_func.counter = 0
				return result
		try_func.counter = 0
		try_func.enabled = enabled

		return try_func

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Parallel(Layer):					# pylint: disable=too-few-public-methods
	""" A container for applying tensor operations to each element in a list.
		This is effectively a map operation, like Theano's `scan` or Keras's
		`TimeDistributed`.

		# Example

		```
		parallel:
		  apply:
		    - dense:
		        size: 512
			- convolution:
			    kernels: 16
				...
		```
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Create a new parallel container.
		"""
		super().__init__(*args, **kwargs)

	###########################################################################
	def _parse(self, engine):
		""" Parse the child containers
		"""
		# Parse self
		if 'apply' not in self.args:
			raise ParsingError('Missing "apply" key in parallel container.')
		target = engine.evaluate(self.args['apply'])

		# Parse children
		for child in target:
			self.new_child_from_data(child).parse(engine)

	###########################################################################
	def _build(self, model):
		""" Instantiate the container.
		"""
		backend = model.get_backend()

		if backend.get_name() == 'pytorch':
			from kur.backend.pytorch.modules import parallel

		for child_index, child in enumerate(self.children):
			for layer_index, layer in enumerate(child.build(model)):

				if backend.get_name() == 'keras':

					import keras.layers as L	# pylint: disable=import-error
					yield L.TimeDistributed(
						layer,
						name='{}_{}_{}_{}'.format(self.name, child.name,
							child_index, layer_index)
					)

				elif backend.get_name() == 'pytorch':

					def connect(inputs, inner_connect=layer, child=child):
						""" Connects the layer
						"""
						operation = inner_connect([{
							'shape' : inputs[0]['shape'][1:],
							'layer' : None
						}])['layer'].op

						return {
							'shape' : (inputs[0]['shape'][0], ) + child.shape(
								(inputs[0]['shape'][1:], )
							),
							'layer' : model.data.add_operation(
								parallel(operation)
							)(inputs[0]['layer'])
						}

					yield connect

				else:
					raise ValueError(
						'Unknown or unsupported backend: {}'.format(backend)
					)

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Parallel layers only take a single input.')
		input_shape = input_shapes[0]

		cur_shape = input_shape[1:]
		for child in self.get_children(recursive=True):
			if child.terminal():
				cur_shape = child.shape([cur_shape])
		return (input_shape[0], ) + cur_shape

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .logger import Logger

###############################################################################
class PersistentLogger(Logger):
	""" A class for storing log data in a fast binary format that can be
		quickly appended to or randomly seeked.
	"""

	###########################################################################
	def get_best_training_loss(self):
		""" Returns the best historical training loss.

			# Return value

			If historical training loss is present in the log, it is returned.
			Otherwise, None is returned.
		"""
		raise NotImplementedError

	###########################################################################
	def get_best_validation_loss(self):
		""" Returns the best historical validation loss.

			# Return value

			If historical validation loss is present in the log, it is returned.
			Otherwise, None is returned.
		"""
		raise NotImplementedError

	###########################################################################
	def enumerate_statistics(self):
		""" Returns the available statistics.
		"""
		raise NotImplementedError

	###########################################################################
	def load_statistic(self, statistic):
		""" Loads a particular statistic.

			# Return value

			A tuple of `(batch, timestamp, value)`. Each entry in the tuple may
			be None. All non-None entries will be numpy arrays of the same
			length, containing the respective data.
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import json
import signal
import atexit
import time
import sys
import argparse
import logging

from . import __version__, __homepage__
from .utils import logcolor
from . import Kurfile
from .engine import JinjaEngine

logger = logging.getLogger(__name__)

###############################################################################
def parse_kurfile(filename, engine):
	""" Parses a Kurfile.

		# Arguments

		filename: str. The path to the Kurfile to load.

		# Return value

		Kurfile instance
	"""
	spec = Kurfile(filename, engine)
	spec.parse()
	return spec

###############################################################################
def dump(args):
	""" Dumps the Kurfile to stdout as a JSON blob.
	"""
	spec = parse_kurfile(args.kurfile, args.engine)
	print(json.dumps(spec.data, sort_keys=True, indent=4))

###############################################################################
def train(args):
	""" Trains a model.
	"""
	spec = parse_kurfile(args.kurfile, args.engine)
	func = spec.get_training_function()
	func(step=args.step)

###############################################################################
def test(args):
	""" Tests a model.
	"""
	spec = parse_kurfile(args.kurfile, args.engine)
	func = spec.get_testing_function()
	func(step=args.step)

###############################################################################
def evaluate(args):
	""" Evaluates a model.
	"""
	spec = parse_kurfile(args.kurfile, args.engine)
	func = spec.get_evaluation_function()
	func(step=args.step)

###############################################################################
def build(args):
	""" Builds a model.
	"""
	spec = parse_kurfile(args.kurfile, args.engine)

	if args.compile == 'auto':
		result = []
		for section in ('train', 'test', 'evaluate'):
			if section in spec.data:
				result.append((section, 'data' in spec.data[section]))
		if not result:
			logger.info('Trying to build a bare model.')
			args.compile = 'none'
		else:
			args.compile, has_data = sorted(result, key=lambda x: not x[1])[0]
			logger.info('Trying to build a "%s" model.', args.compile)
			if not has_data:
				logger.info('There is not data defined for this model, '
					'though, so we will be running as if --bare was '
					'specified.')
	elif args.compile == 'none':
		logger.info('Trying to build a bare model.')
	else:
		logger.info('Trying to build a "%s" model.', args.compile)

	if args.bare or args.compile == 'none':
		provider = None
	else:
		provider = spec.get_provider(args.compile)

	spec.get_model(provider)

	if args.compile == 'none':
		return
	elif args.compile == 'train':
		target = spec.get_trainer(with_optimizer=True)
	elif args.compile == 'test':
		target = spec.get_trainer(with_optimizer=False)
	elif args.compile == 'evaluate':
		target = spec.get_evaluator()
	else:
		logger.error('Unhandled compilation target: %s. This is a bug.',
			args.compile)
		return 1

	target.compile()

###############################################################################
def prepare_data(args):
	""" Prepares a model's data provider.
	"""
	spec = parse_kurfile(args.kurfile, args.engine)

	if args.target == 'auto':
		result = None
		for section in ('train', 'validate', 'test', 'evaluate'):
			if section in spec.data and 'data' in spec.data[section]:
				result = section
				break
		if result is None:
			raise ValueError('No data sections were found in the Kurfile.')
		args.target = result

	logger.info('Preparing data sources for: %s', args.target)

	provider = spec.get_provider(args.target)

	if args.assemble:

		spec.get_model(provider)

		if args.target == 'train':
			target = spec.get_trainer(with_optimizer=True)
		elif args.target == 'test':
			target = spec.get_trainer(with_optimizer=False)
		elif args.target == 'evaluate':
			target = spec.get_evaluator()
		else:
			logger.error('Unhandled assembly target: %s. This is a bug.',
				args.target)
			return 1

		target.compile(assemble_only=True)

	batch = None
	for batch in provider:
		break
	if batch is None:
		logger.error('No batches were produced.')
		return 1

	num_entries = None
	keys = sorted(batch.keys())
	num_entries = len(batch[keys[0]])
	for entry in range(num_entries):
		print('Entry {}/{}:'.format(entry+1, num_entries))
		for k in keys:
			print('  {}: {}'.format(k, batch[k][entry]))

	if num_entries is None:
		logger.error('No data sources was produced.')
		return 1

###############################################################################
def version(args):							# pylint: disable=unused-argument
	""" Prints the Kur version and exits.
	"""
	print('Kur, by Deepgram -- deep learning made easy')
	print('Version: {}'.format(__version__))
	print('Homepage: {}'.format(__homepage__))

###############################################################################
def do_monitor(args):
	""" Handle "monitor" mode.
	"""

	# If we aren't running in monitor mode, then we are done.
	if not args.monitor:
		return

	# This is the main retry loop.
	while True:
		# Fork the process.
		logger.info('Forking child process.')
		pid = os.fork()

		# If we are the child, leave this function and work.
		if pid == 0:
			logger.info('We are a newly spawned child process.')
			return

		logger.info('Child process spawned: %d', pid)

		# Wait for the child to die. If we die first, kill the child.
		atexit.register(kill_process, pid)
		try:
			_, exit_status = os.waitpid(pid, 0)
		except KeyboardInterrupt:
			break
		atexit.unregister(kill_process)

		# Process the exit code.
		signal_number = exit_status & 0xFF
		exit_code = (exit_status >> 8) & 0xFF
		core_dump = bool(0x80 & signal_number)

		if signal_number == 0:
			logger.info('Child process exited with exit code: %d.', exit_code)
		else:
			logger.info('Child process exited with signal %d (core dump: %s).',
				signal_number, core_dump)

		retry = False
		if os.WIFSIGNALED(exit_status):
			if os.WTERMSIG(exit_status) == signal.SIGSEGV:
				logger.error('Child process seg faulted.')
				retry = True

		if not retry:
			break

	sys.exit(0)

###############################################################################
def kill_process(pid):
	""" Kills a child process by PID.
	"""

	# Maximum time we wait (in seconds) before we send SIGKILL.
	max_timeout = 60

	# Terminate child process
	logger.info('Sending Ctrl+C to the child process %d', pid)
	os.kill(pid, signal.SIGINT)

	start = time.time()
	while True:
		now = time.time()

		# Check the result.
		result = os.waitpid(pid, os.WNOHANG)
		if result != (0, 0):
			# The child process is dead.
			break

		# Check the timeout.
		if now - start > max_timeout:
			# We've waited too long.
			os.kill(pid, signal.SIGKILL)
			break

		# Keep waiting.
		logger.debug('Waiting patiently...')
		time.sleep(0.5)

###############################################################################
def parse_args():
	""" Constructs an argument parser and returns the parsed arguments.
	"""
	parser = argparse.ArgumentParser(
		description='Descriptive deep learning')
	parser.add_argument('--no-color', action='store_true',
		help='Disable colorful logging.')
	parser.add_argument('-v', '--verbose', default=0, action='count',
		help='Increase verbosity. Can be specified twice for debug-level '
			'output.')
	parser.add_argument('--monitor', action='store_true',
		help='Run Kur in monitor mode, which tries to recover from critical '
			'errors, like segmentation faults.')
	parser.add_argument('--version', action='store_true',
		help='Display version and exit.')

	subparsers = parser.add_subparsers(dest='cmd', help='Sub-command help.')

	subparser = subparsers.add_parser('train', help='Trains a model.')
	subparser.add_argument('--step', action='store_true',
		help='Interactive debug; prompt user before submitting each batch.')
	subparser.add_argument('kurfile', help='The Kurfile to use.')
	subparser.set_defaults(func=train)

	subparser = subparsers.add_parser('test', help='Tests a model.')
	subparser.add_argument('--step', action='store_true',
		help='Interactive debug; prompt user before submitting each batch.')
	subparser.add_argument('kurfile', help='The Kurfile to use.')
	subparser.set_defaults(func=test)

	subparser = subparsers.add_parser('evaluate', help='Evaluates a model.')
	subparser.add_argument('--step', action='store_true',
		help='Interactive debug; prompt user before submitting each batch.')
	subparser.add_argument('kurfile', help='The Kurfile to use.')
	subparser.set_defaults(func=evaluate)

	subparser = subparsers.add_parser('build',
		help='Tries to build a model. This is useful for debugging a model.')
	subparser.add_argument('kurfile', help='The Kurfile to use.')
	subparser.add_argument('-c', '--compile',
		choices=['none', 'train', 'test', 'evaluate', 'auto'], default='auto',
		help='Try to compile the specified variation of the model. If '
			'--compile=none, then it only tries to assemble the model, not '
			'compile anything. --compile=none implies --bare')
	subparser.add_argument('-b', '--bare', action='store_true',
		help='Do not attempt to load the data providers. In order for your '
			'model to build correctly with this option, you will need to '
			'specify shapes for all of your inputs.')
	subparser.set_defaults(func=build)

	subparser = subparsers.add_parser('dump',
		help='Dumps the Kurfile out as a JSON blob. Useful for debugging.')
	subparser.add_argument('kurfile', help='The Kurfile to use.')
	subparser.set_defaults(func=dump)

	subparser = subparsers.add_parser('data',
		help='Does not actually compile anything, but only prints out a '
			'single batch of data. This is useful for debugging data sources.')
	subparser.add_argument('kurfile', help='The Kurfile to use.')
	subparser.add_argument('-t', '--target',
		choices=['train', 'validate', 'test', 'evaluate', 'auto'],
		default='auto', help='Try to produce data corresponding to a specific '
			'variation of the model.')
	subparser.add_argument('--assemble', action='store_true', help='Also '
		'begin assembling the model to pull in compile-time, auxiliary data '
		'sources.')
	subparser.set_defaults(func=prepare_data)

	return parser.parse_args()

###############################################################################
def main():
	""" Entry point for the Kur command-line script.
	"""
	args = parse_args()

	loglevel = {
		0 : logging.WARNING,
		1 : logging.INFO,
		2 : logging.DEBUG
	}
	config = logging.basicConfig if args.no_color else logcolor.basicConfig
	config(
		level=loglevel.get(args.verbose, logging.DEBUG),
		format='{color}[%(levelname)s %(asctime)s %(name)s:%(lineno)s]{reset} '
			'%(message)s'.format(
				color='' if args.no_color else '$COLOR',
				reset='' if args.no_color else '$RESET'
			)
	)
	logging.captureWarnings(True)

	do_monitor(args)

	if args.version:
		args.func = version
	elif not hasattr(args, 'func'):
		print('Nothing to do!', file=sys.stderr)
		print('For usage information, try: kur --help', file=sys.stderr)
		print('Or visit our homepage: {}'.format(__homepage__))
		sys.exit(1)

	engine = JinjaEngine()
	setattr(args, 'engine', engine)

	sys.exit(args.func(args) or 0)

###############################################################################
if __name__ == '__main__':
	main()

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy

from kur.sources import VanillaSource, StackSource

###############################################################################
class TestStack:
	""" Tests for the StackSource.
	"""

	###########################################################################
	def test_single_fixed(self):
		""" Test a known permutation with exactly one source.
		"""
		data = numpy.array([1, 2, 3, 4, 5])
		indices = [4, 2, 1, 3, 0]
		expected = numpy.array([5, 3, 2, 4, 1])

		source = VanillaSource(data)
		stack = StackSource(source)
		stack.shuffle(indices)
		result = numpy.array([actual for batch in stack for actual in batch])

		assert numpy.allclose(result, expected)

	###########################################################################
	def test_single(self):
		""" Test a random permutation with exactly one source.
		"""
		data = numpy.array([1, 2, 3, 4, 5])
		source = VanillaSource(data)
		stack = StackSource(source)

		assert len(stack) == len(data)
		assert stack.shape() == ()

		indices = numpy.random.permutation(len(data))

		preshuffled = numpy.copy(data)
		stack.shuffle(indices)

		for i, x in enumerate(data):
			assert preshuffled[indices[i]] == x

		cur = 0
		for batch in stack:
			for actual in batch:
				expected = data[cur]
				assert actual == expected
				cur += 1

		assert cur == len(data)

	###########################################################################
	def test_double(self):
		""" Test a random permutation with exactly two sources.
		"""
		data = [
			numpy.array([1, 2, 3, 4, 5]),
			numpy.array([6, 7, 8, 9, 0])
		]
		sources = [
			VanillaSource(data[0]),
			VanillaSource(data[1])
		]
		stack = StackSource(*sources)

		total = sum(len(x) for x in data)
		assert len(stack) == total
		assert stack.shape() == ()

		indices = numpy.random.permutation(total)

		preshuffled = [numpy.copy(x) for x in data]
		stack.shuffle(indices)

		sub = [i for i in indices if i < len(data[0])]
		for i, x in enumerate(data[0]):
			assert preshuffled[0][sub[i]] == x

		sub = [i % len(data[0]) for i in indices if i >= len(data[0])]
		for i, x in enumerate(data[1]):
			assert preshuffled[1][sub[i]] == x

		cur = 0
		for batch in stack:
			for actual in batch:

				i = indices[cur]
				expected = preshuffled[i//5][i%5]
				assert actual == expected
				cur += 1

		assert cur == total

	###########################################################################
	def test_triple_fixed(self):
		""" Test a known permutation with exactly three sources.
		"""
		data = [
			numpy.arange(10, 16),
			numpy.arange(16, 20),
			numpy.arange(20, 23)
		]
		indices = [12, 3, 5, 6, 11, 0, 10, 2, 1, 4, 9, 7, 8]
		expected = numpy.array(
			[22, 13, 15, 16, 21, 10, 20, 12, 11, 14, 19, 17, 18]
		)

		sources = [VanillaSource(x) for x in data]
		stack = StackSource(*sources)
		stack.shuffle(indices)
		result = numpy.array([actual for batch in stack for actual in batch])

		assert numpy.allclose(result, expected)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Operator

###############################################################################
class Meta(Operator):
	""" A meta container is a placeholder for user-defined containers, also
		called templates, which act as macros.
	"""

	###########################################################################
	def is_anonymous(self):
		""" This container's "name" is dynamic, so we need to act as if we have
			no arguments.
		"""
		return True

	###########################################################################
	def _resolve(self, engine):
		""" Figure out which template we represent.
		"""
		name = None
		template = None
		for key in self.data:
			candidate = engine.get_template(key)
			if candidate is None:
				continue
			if template is not None:
				raise ValueError('Ambiguous template substitution. Could be '
					'either "{}" or "{}".'.format(key, name))
			template = candidate
			name = key

		if template is None:
			raise ValueError('Failed to find an appropriate template for '
				'container: {}'.format(self.data))

		return name, template

	###########################################################################
	def _parse(self, engine):
		""" Parses out the meta-container.
		"""
		# Figure out which template we represent.
		name, template = self._resolve(engine)
		self.args = engine.evaluate(self.data[name])

		if self.inputs:
			inputs = self.inputs
		else:
			inputs = engine.state.layers[-1]

		# Put the arguments in scope for the children.
		with engine.scope(args=self.args, inputs=inputs):
			with engine.scope(**self.args):

				# Parse every child.
				for entry in template:
					entry = engine.evaluate(entry)
					self.new_child_from_data(entry).parse(engine)

	###########################################################################
	def _build(self, model):
		""" Construct each child.
		"""
		for child in self.children:
			yield from child.build(model)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import csv

import numpy

from . import Supplier
from ..sources import VanillaSource
from ..utils import package

logger = logging.getLogger(__name__)

###############################################################################
class CsvSupplier(Supplier):
	""" A supplier which parses a CSV file.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the supplier.
		"""
		return 'csv'

	###########################################################################
	def __init__(self, path=None, url=None, checksum=None, format=None,
		header=None, *args, **kwargs):
		""" Creates a new Pickle supplier.

			# Arguments

			source: str. Filename of pickled file to load.
		"""
		super().__init__(*args, **kwargs)

		target = {
			'path' : path,
			'url' : url,
			'checksum' : checksum,
			'header' : header,
			'format' : format
		}
		self.parse_source({k : v for k, v in target.items() if v is not None})
		self.data = None

	###########################################################################
	def parse_source(self, target):
		""" Parses the data specification.
		"""
		if isinstance(target, str):
			target = {'path' : target}
		path, _ = package.install(
			url=target.get('url'),
			path=target.get('path'),
			checksum=target.get('checksum')
		)

		self.format = target.get('format', {})
		self.header = target.get('header', True)

		self.source = path

	###########################################################################
	def _load(self):
		""" Loads the data (only if it hasn't already been loaded).

			# Note

			By the end of this call, `self.data` will be a dictionary with keys
			for each CSV column, and values which are Source instances for each
			column (e.g., VanillaSource wrapping a numpy array).
		"""
		if self.data is not None:
			return

		logger.info('Reading in CSV file: %s', self.source)
		with open(self.source, newline='') as fh:
			kwargs = {}

			try:
				# Deduce the file format.
				kwargs['dialect'] = csv.Sniffer().sniff(fh.read(1024))
			except csv.Error:
				logger.warning('Failed to sniff data format for file: %s. You '
					'might need to manually choose delimiters, quote '
					'characters, etc.', self.source)
			finally:
				fh.seek(0)

			# Override the format, if necessary.
			if 'delimiter' in self.format:
				kwargs['delimiter'] = self.format['delimiter']
			if 'quote' in self.format:
				kwargs['quotechar'] = self.format['quote']

			# Prepare the reader.
			reader = csv.reader(fh, **kwargs)

			# If we have a header to read, read it now.
			if self.header:
				for row in reader:
					header = row
					break

			# Read the data
			data = numpy.asarray(list(reader), dtype='float32')

			# If we did not have a header, then now is a good time to figure
			# out what our auto-generated header names are going to be.
			if not self.header:
				header = ['column_{}'.format(i) for i in range(len(data[0]))]

		logger.info('Finished reading CSV file.')

		data = data.T
		self.data = {k : VanillaSource(v) for k, v in zip(header, data)}

	###########################################################################
	def get_sources(self, sources=None):
		""" Returns all sources from this provider.
		"""
		self._load()

		if sources is None:
			sources = list(self.data.keys())
		elif not isinstance(sources, (list, tuple)):
			sources = [sources]

		for source in sources:
			if source not in self.data:
				raise KeyError(
					'Invalid data key: {}. Valid keys are: {}'.format(
						source, ', '.join(str(k) for k in self.data)
				))

		return {k : self.data[k] for k in sources}

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import textwrap
import pickle
import os
import tempfile

import yaml
import pytest
import numpy

from kur.kurfile import Kurfile

from modkurfile import modify_kurfile
from example import example

@pytest.fixture(scope='session')
def tutorial_yaml():
	return textwrap.dedent("""
		model:
		  - input: point
		  - dense: 128
		  - activation: tanh
		  - dense: 1
		  - activation: tanh
		    name: above

		train:
		  data:
		    - pickle: train.pkl
		  epochs: 10
		  weights: best.w
		  log: tutorial-log

		validate:
		  data:
		    - pickle: validate.pkl
		  weights: best.w

		test:
		  data:
		    - pickle: test.pkl
		  weights: best.w

		evaluate:
		  data:
		    - pickle: evaluate.pkl
		  weights: best.w
		  destination: output.pkl

		loss:
		  - target: above
		    name: mean_squared_error
	""")

@pytest.fixture(scope='session')
def num_samples():
	return 10

def make_points(num_samples, filename):
	x = numpy.array([
		numpy.random.uniform(-numpy.pi, numpy.pi, num_samples),
		numpy.random.uniform(-1, 1, num_samples)
	]).T
	y = (numpy.sin(x[:,0]) < x[:,1]).astype(numpy.float32) * 2 - 1
	with open(filename, 'wb') as fh:
		fh.write(pickle.dumps({'point' : x, 'above' : y}))

@pytest.fixture(scope='session')
def kurfile(tutorial_yaml):
	result = Kurfile(yaml.load(tutorial_yaml))

	modify_kurfile(result.data)
	for k in ('train', 'validate', 'test', 'evaluate'):
		if k in result.data and 'data' in result.data[k]:
			result.data[k]['data'][0]['pickle'] = os.path.join(
				tempfile.gettempdir(),
				result.data[k]['data'][0]['pickle']
			)

	result.parse()
	return result

@pytest.fixture(scope='session')
def tutorial_data(request, kurfile, num_samples):
	files = []
	for k in ('train', 'validate', 'test', 'evaluate'):
		if k not in kurfile.data:
			continue
		if 'data' not in kurfile.data[k]:
			continue
		if not kurfile.data[k]['data']:
			continue
		files.append(kurfile.data[k]['data'][0]['pickle'])

	for filename in files:
		make_points(num_samples, filename)

	yield

	for filename in files:
		os.unlink(filename)

class TestTutorial:

	@example
	def test_train(self, kurfile, tutorial_data):
		kurfile.get_training_function()()

	@example
	def test_test(self, kurfile, tutorial_data):
		kurfile.get_testing_function()()

	@example
	def test_evaluate(self, kurfile, tutorial_data):
		kurfile.get_evaluation_function()()

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest
import numpy

from kur.sources import VanillaSource, StackSource, ChunkSource

###############################################################################
def flatten_source(source):
	""" Flattens a Source into a single numpy array.
	"""
	data = None
	start = 0
	for batch in source:
		if data is None:
			data = numpy.empty(
				(len(source), ) + source.shape(),
				dtype=batch.dtype
			)
		data[start:start+len(batch)] = batch[:]
		start += len(batch)

	return data

###############################################################################
@pytest.fixture(
	params=[20, 100, int(1.2 * ChunkSource.DEFAULT_CHUNK_SIZE)]
)
def num_entries(request):
	""" How many entries per Source
	"""
	return request.param

###############################################################################
@pytest.fixture(
	params=[None, 7]
)
def chunk_size(request):
	""" Set the chunk size for the source (or None to use the default).
	"""
	return request.param

###############################################################################
@pytest.fixture(
	params=[0, 1, 2]
)
def source(request, num_entries, chunk_size):
	""" Creates the data Source to test.
	"""
	num_stacks = request.param
	if num_stacks == 0:
		result = VanillaSource(numpy.random.permutation(num_entries)+10)
	else:
		result = StackSource(*[
			VanillaSource(
				numpy.random.permutation(num_entries) + 100*i
			) for i in range(num_stacks)
		])
	if chunk_size is not None:
		result.set_chunk_size(chunk_size)
	return result

###############################################################################
class TestShuffling:

	###########################################################################
	def test_shuffle(self, source):

		original_data = flatten_source(source)
		assert numpy.allclose(original_data, flatten_source(source))

		num_entries = len(source)
		current_permutation = numpy.arange(num_entries)

		for _ in range(5):
			indices = numpy.random.permutation(num_entries)
			source.shuffle(indices)

			current_permutation = current_permutation[indices]
			assert numpy.allclose(
				original_data[current_permutation],
				flatten_source(source)
			)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import threading
import logging
import ctypes
from functools import wraps, partial

logger = logging.getLogger(__name__)

###########################################################################
def locked(lock):
	""" Wrap function in a thread-safe lock. This is similar to Java's
		`synchronized(this)`.
	"""

	###########################################################################
	def wrapper(func):
		""" Closure for the actual decorator.
		"""

		#######################################################################
		@wraps(func)
		def wrapped(*args, **kwargs):
			""" The wrapped (locked) function call.
			"""
			with lock:
				func(*args, **kwargs)
		return wrapped

	return wrapper

###############################################################################
class CudaError(Exception):
	""" Exception class for CUDA errors.
	"""
	pass

###############################################################################
# The lock
_LOCK = threading.RLock()

###############################################################################
def ready(func):
	""" Decorator for a class instance method which requires that the class's
		"_ready" property be set.
	"""

	###########################################################################
	@wraps(func)
	def wrapper(self, *args, **kwargs):
		""" The wrapped function.
		"""
		if not self._ready:
			raise ValueError('CUDA has not been initialized for this object. '
				'Be sure to only use this object only within Python context '
				'management.')
		return func(self, *args, **kwargs)

	return wrapper

###############################################################################
class DeviceHandleRaw(ctypes.Structure):
	pass
DeviceHandle = ctypes.POINTER(DeviceHandleRaw)

###############################################################################
class DeviceMemory(ctypes.Structure):
	_fields_ = [
		('total', ctypes.c_ulonglong),
		('free', ctypes.c_ulonglong),
		('used', ctypes.c_ulonglong)
	]

###############################################################################
class DeviceUtilization(ctypes.Structure):
	_fields_ = [
		('gpu', ctypes.c_uint),
		('memory', ctypes.c_uint),
	]

################################################################################
class Device:
	""" Manages individual GPU devices.

		# References

		- http://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html
	"""

	###########################################################################
	def __init__(self, handle, context):
		self.handle = handle
		self.context = context

	###########################################################################
	def __str__(self):
		return 'Device(index="{}", uuid="{}")' \
			.format(self.index(), self.uuid())

	###########################################################################
	def __repr__(self):
		return str(self)

	###########################################################################
	def __getattr__(self, name):
		if hasattr(self.context, name):
			attr = getattr(self.context, name)
			if callable(attr) and hasattr(attr, '_device') and attr._device:
				return partial(attr, self)
		raise AttributeError('No such attribute for a device: {}'.format(name))

###############################################################################
def mark_device(func):
	func._device = True
	return func

###############################################################################
class CudaContext:

	NVIDIA_ML = 'libnvidia-ml.so.1'
	_dll = None
	_ref = 0
	_ptr = {}

	###########################################################################
	def __init__(self):
		self._ready = False
		self._devices = None

	###########################################################################
	def __enter__(self):
		self._init()
		self._devices = {}
		self._ready = True
		return self

	###########################################################################
	def __exit__(self, exc_type, exc_val, exc_tb):
		self._ready = False
		self._del()

	###########################################################################
	@classmethod
	@locked(_LOCK)
	def _init(cls):
		if cls._dll is None:
			logger.debug('Loading NVIDIA ML library.')
			try:
				cls._dll = ctypes.CDLL(cls.NVIDIA_ML)
			except OSError:
				raise CudaError('Failed to find NVIDIA ML library.')

		if not cls._ref:
			logger.debug('Initializing NVIDIA ML.')
			init = cls._get_ptr('nvmlInit_v2')
			if init():
				raise CudaError('Failed to initialize NVIDIA ML.')

		cls._ref += 1

	###########################################################################
	@classmethod
	def _get_ptr(cls, function_name):
		result = cls._ptr.get(function_name)
		if result is None:
			try:
				result = getattr(cls._dll, function_name)
			except AttributeError:
				logger.exception('No such function found in the NVIDIA ML '
					'library: %s')
				raise
			cls._ptr[function_name] = result
		return result

	###########################################################################
	@classmethod
	@locked(_LOCK)
	def _del(cls):
		cls._ref -= 1
		if not cls._ref:
			logger.debug('Shutting down NVIDIA ML.')
			shutdown = cls._get_ptr('nvmlShutdown')
			if shutdown():
				raise CudaError('Failed to shutdown NVIDIA ML.')

	###########################################################################
	def __iter__(self):
		""" Iterates over devices.
		"""
		return (self.device(i) for i in range(len(self)))

	###########################################################################
	@ready
	def __len__(self):
		""" Retrieves the number of GPU devices.
		"""
		result = ctypes.c_uint()
		func = self._get_ptr('nvmlDeviceGetCount_v2')
		if func(ctypes.byref(result)):
			raise CudaError('Failed to shutdown NVIDIA ML.')
		return result.value

	###########################################################################
	@ready
	def _get_handle_by_index(self, index):
		c_index = ctypes.c_uint(index)
		handle = DeviceHandle()
		func = self._get_ptr('nvmlDeviceGetHandleByIndex_v2')
		if func(c_index, ctypes.byref(handle)):
			raise CudaError('Failed to get device by index: {}'.format(index))
		return handle

	###########################################################################
	@ready
	def _get_handle_by_uuid(self, uuid):
		c_uuid = ctypes.c_char_p(uuid)
		handle = DeviceHandle()
		func = self._get_ptr('nvmlDeviceGetHandleByUUID')
		if func(c_uuid, ctypes.byref(handle)):
			raise CudaError('Failed to get device by UUID: {}'.format(uuid))
		return handle

	###########################################################################
	def __getitem__(self, device):
		""" A convenience function for retrieving devices by index.
		"""
		return self.device(device)

	###########################################################################
	def device(self, device):
		""" Retrieves a Device handler.

			# Arguments

			device: int or str. If an integer, it is the zero-based index of
				the device to retrieve. If a string, it is the UUID of the
				device to retrieve.

			# Return value

			A Device instance. If no such device can be found, an exception is
			raised.
		"""
		if isinstance(device, int):
			if device < 0 or device >= len(self):
				raise ValueError('Out-of-range device index: {}'
					.format(device))
			if device not in self._devices:
				handle = Device(self._get_handle_by_index(device), self)
				self._devices[device] = handle
			else:
				handle = self._devices[device]
		elif isinstance(device, (str, bytes)):
			if isinstance(device, str):
				device = device.encode('utf-8')
			handle = Device(self._get_handle_by_uuid(device), self)
			index = handle.index()
			if index not in self._devices:
				self._devices[index] = handle
			else:
				# Don't accidentally create additional handles that you give to
				# the user.
				handle = self._devices[index]
		else:
			raise TypeError(
				'Index must be a integer index, or a str/bytes UUID.')

		return handle

	###########################################################################
	@mark_device
	@ready
	def name(self, device):
		""" Returns the name of the device as a string.
		"""
		logger.debug('Calculating device name.')
		buffer_size = 64
		name = ctypes.create_string_buffer(buffer_size)
		func = self._get_ptr('nvmlDeviceGetName')
		if func(device.handle, name, ctypes.c_uint(buffer_size)):
			raise CudaError('Failed to get device name.')
		return name.value.decode('utf-8')

	###########################################################################
	@mark_device
	@ready
	def uuid(self, device):
		""" Returns the device UUID as a string.
		"""
		logger.debug('Calculating device UUID.')
		buffer_size = 80
		uuid = ctypes.create_string_buffer(buffer_size)
		func = self._get_ptr('nvmlDeviceGetUUID')
		if func(device.handle, uuid, ctypes.c_uint(buffer_size)):
			raise CudaError('Failed to get UUID for device.')
		return uuid.value.decode('utf-8')

	###########################################################################
	@mark_device
	@ready
	def index(self, device):
		""" Returns the zero-based index of the device.
		"""
		logger.debug('Calculating device index.')
		index = ctypes.c_uint()
		func = self._get_ptr('nvmlDeviceGetIndex')
		if func(device.handle, ctypes.byref(index)):
			raise CudaError('Failed to get index for device.')
		return index.value

	###########################################################################
	@mark_device
	def kernel_utilization(self, device):
		""" Returns the kernel utilization of the device, as an integer between
			0 (not utilized) and 100 (fully utilized).
		"""
		return self._utilization(device).gpu

	###########################################################################
	@mark_device
	def memory_utilization(self, device):
		""" Returns the memory utilization of the device, as an integer between
			0 (not utilized) and 100 (fully utilized).
		"""
		return self._utilization(device).memory

	###########################################################################
	@ready
	def _utilization(self, device):
		logger.debug('Calculating device utilization.')
		result = DeviceUtilization()
		func = self._get_ptr('nvmlDeviceGetUtilizationRates')
		if func(device.handle, ctypes.byref(result)):
			raise CudaError('Failed to get device utilization.')
		return result

	###########################################################################
	@mark_device
	def used_memory(self, device):
		""" Returns the used memory on the device, in MiB.
		"""
		return self._memory(device).used / 2**20

	###########################################################################
	@mark_device
	def free_memory(self, device):
		""" Returns the free memory on the device, in MiB.
		"""
		return self._memory(device).free / 2**20

	###########################################################################
	@mark_device
	def total_memory(self, device):
		""" Returns the total memory on the device, in MiB.
		"""
		return self._memory(device).total / 2**20

	###########################################################################
	@ready
	def _memory(self, device):
		logger.debug('Calculating device memory footprint.')
		result = DeviceMemory()
		func = self._get_ptr('nvmlDeviceGetMemoryInfo')
		if func(device.handle, ctypes.byref(result)):
			raise CudaError('Failed to get memory utilization.')
		return result

	###########################################################################
	@mark_device
	def business(self, device):
		""" Calculates how busy a device is ("busy-ness").

			# Return value

			An integer between 0 (idle) and 100 (fully loaded) indicating how
			busy the device is.
		"""
		logger.debug('Calculating device business...')
		memory = self._memory(device)
		utilization = self._utilization(device)
		x = (
			100 * memory.used / memory.total,
			utilization.gpu,
			utilization.memory
		)
		return int(sum(x) / len(x))

	###########################################################################
	@mark_device
	def is_busy(self, device):
		""" Convenience function for checking if a device is in use.
		"""
		return self.business(device) >= 5

	###########################################################################
	def rank_available(self, num=None):
		""" Returns devices in the order of increasing usage.

			# Arguments

			num: int or None. If None, returns all devices. If an integer, then
				returns up to `num` devices.d$x t

			# Return value

			A list of devices, ordered by increasing usage (so that the idling
			devices would come first). If two devices tie for usage, then the
			device with the lower index is ordered first.
		"""
		logger.debug('Calculating device availability...')
		if num is not None and (num <= 0 or num > len(self)):
			raise IndexError('Invalid number of instances to return. Use None '
				'to return all available.')

		devices = sorted(
			self,
			key=lambda device: (device.business(), device.index())
		)
		if not devices:
			raise ValueError('No GPUs found.')

		if num is None:
			return devices
		else:
			return devices[:num]

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import warnings
from . import Layer

logger = logging.getLogger(__name__)

###############################################################################
class Output(Layer):				# pylint: disable=too-few-public-methods
	""" A named dummy layer for outputs.
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new placeholder.
		"""
		super().__init__(*args, **kwargs)

	###########################################################################
	def _parse_pre(self, engine):
		""" Pre-parsing hook.
		"""
		super()._parse_pre(engine)

		container_name = self.get_container_name()
		if container_name in self.data:
			data = engine.evaluate(self.data[container_name])
			if isinstance(data, str):
				if 'name' in self.data:
					name = engine.evaluate(self.data['name'])
					logger.warning('Conflicting naming schemes for '
						'output layer: "%s" and "%s". Using: "%s".',
						data, name, name)
				else:
					logger.debug('Using short-hand name for output: %s', data)
					self.name = data
		else:
			warnings.warn('Parsing oddity in output: {}'.format(self.data))

	###########################################################################
	def _parse_post(self, engine):
		""" Post-parsing hook.
		"""
		self.sink = True
		super()._parse_post(engine)

	###########################################################################
	def _build(self, model):
		""" Create the backend-specific placeholder.
		"""
		yield None
		"""backend = model.get_backend()
		if backend.get_name() == 'keras':
			yield lambda inputs: inputs[0]
		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))"""

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Outputs only take a single input.')
		input_shape = input_shapes[0]
		return input_shape

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from kur.utils import get_subclasses

###############################################################################
class EvaluationHook:					# pylint: disable=too-few-public-methods
	""" Base class for all evaluation hooks.

		Evaluation hooks are post-evaluation callbacks that can take a crack at
		modifying or reviewing the evaluation results.
	"""

	###########################################################################
	@staticmethod
	def from_specification(spec):
		""" Creates a new evaluation hook from a specification.

			# Arguments

			spec: str or dict. The specification object that was given.

			# Return value

			New EvaluationHook instance.
		"""
		if isinstance(spec, str):
			name = spec
			target = EvaluationHook.get_hook_by_name(spec)
			params = {}
			meta = {}
		elif isinstance(spec, dict):
			candidates = {}
			for key in spec:
				try:
					cls = EvaluationHook.get_hook_by_name(key)
					candidates[key] = cls
				except ValueError:
					pass

			if not candidates:
				raise ValueError('No valid evaluation hook found in '
					'Kurfile: {}'.format(spec))
			elif len(candidates) > 1:
				raise ValueError('Too many possible evaluation hooks found in '
					'Kurfile: {}'.format(spec))
			else:
				name, target = candidates.popitem()
				meta = dict(spec)
				meta.pop(name)
				params = spec[name] or {}

		else:
			raise ValueError('Expected the evaluation hooks to be a string or '
				'dictionary. We got this instead: {}'.format(spec))

		# At this point:
		# - name: the name of the evaluation hook, as specified in the spec.
		# - target: the class of the evaluation hook
		# - meta: dictionary of "other" keys in the spec that sit alongside the
		#   evaluation hook specification.
		# - params: parameters to be passed to the evaluation hook.

		return target(**params)

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the evaluation hook.

			# Return value

			A lower-case string unique to this evaluation hook.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_hooks():
		""" Returns an iterator to the names of all evaluation hooks.
		"""
		for cls in get_subclasses(EvaluationHook):
			yield cls

	###########################################################################
	@staticmethod
	def get_hook_by_name(name):
		""" Finds a evaluation hook class with the given name.
		"""
		name = name.lower()
		for cls in EvaluationHook.get_all_hooks():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such evaluation hook with name "{}"'.format(name))

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new evaluation hook.
		"""
		if args or kwargs:
			raise ValueError('One or more unexpected or unsupported arguments '
				'to the evaluation hook: {}'.format(
					', '.join(args + list(kwargs.keys()))
			))

	###########################################################################
	def apply(self, current, original, model=None):
		""" Applies the hook to the data.

			# Arguments

			current: tuple. A 2-tuple, `(prediction, truth)`. `prediction` is
				a dictionary whose keys are the names of the output layers
				in the model, and whose values are a list of output tensors.
				If `truth` is not None, then it is a dictionary with the same
				keys as `prediction` which contains ground truth data.
			original: tuple. Same format as `current`. See Notes for details.
				This should not be modified.

			# Return value

			If the hook wants to modify the truth value that the next hook in
			the processing stack sees, then it should return a tuple of
			`(new_prediction, new_truth)` whose values are dictionaries in the
			same form as `data`. If the hook does not modify the truth data,
			then it can also simply return `new_prediction`.

			# Notes

			Hooks can be stacked. The value of `current` is the output of the
			most recent hook in the stack. `original` is the output of the
			model before being passed to the hooks. For the first hook, it is
			true that `output is current`.
		"""
		raise NotImplementedError

###############################################################################
class TrainingHook:					# pylint: disable=too-few-public-methods
	""" Base class for all training hooks.

		Training hooks are between-epoch callbacks that can be used to notify
		external systems about progress.
	"""

	TRAINING_START = object()
	TRAINING_END = object()
	VALIDATION_END = object()
	EPOCH_START = object()
	EPOCH_END = object()

	###########################################################################
	@staticmethod
	def from_specification(spec):
		""" Creates a new training hook from a specification.

			# Arguments

			spec: str or dict. The specification object that was given.

			# Return value

			New TrainingHook instance.
		"""
		if isinstance(spec, str):
			name = spec
			target = TrainingHook.get_hook_by_name(spec)
			params = {}
			meta = {}
		elif isinstance(spec, dict):
			candidates = {}
			for key in spec:
				try:
					cls = TrainingHook.get_hook_by_name(key)
					candidates[key] = cls
				except ValueError:
					pass

			if not candidates:
				raise ValueError('No valid training hook found in '
					'Kurfile: {}'.format(spec))
			elif len(candidates) > 1:
				raise ValueError('Too many possible training hooks found in '
					'Kurfile: {}'.format(spec))
			else:
				name, target = candidates.popitem()
				meta = dict(spec)
				meta.pop(name)
				params = spec[name] or {}

		else:
			raise ValueError('Expected the training hooks to be a string or '
				'dictionary. We got this instead: {}'.format(spec))

		if isinstance(params, dict):
			return target(**params)
		else:
			return target(params)

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the training hook.

			# Return value

			A lower-case string unique to this training hook.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_hooks():
		""" Returns an iterator to the names of all training hooks.
		"""
		for cls in get_subclasses(TrainingHook):
			yield cls

	###########################################################################
	@staticmethod
	def get_hook_by_name(name):
		""" Finds a training hook class with the given name.
		"""
		name = name.lower()
		for cls in TrainingHook.get_all_hooks():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such training hook with name "{}"'.format(name))

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new training hook.
		"""
		if args or kwargs:
			raise ValueError('One or more unexpected or unsupported arguments '
				'to the training hook: {}'.format(
					', '.join(args + list(kwargs.keys()))
			))

	###########################################################################
	def notify(self, status, log=None, info=None):
		""" Notifies the hook that the epoch has ended.

			# Arguments

			status: object. One of the class objects (e.g., `TRAINING_START`)
				that indicates what is going on right now.
			log: Logger instance or None. The logger that has been tracking
				training and validation statistics.
			info: dict or None. Additional information to pass to the hook.
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import warnings

import yaml
import numpy

logger = logging.getLogger(__name__)

###############################################################################
class Normalize:
	""" Class for normalizing data (centering, whitening, re-scaling, etc.)
	"""

	APPROXIMATION_CUTOFF = 10000

	###########################################################################
	def __init__(self, center=True, scale=True, rotate='zca'):
		""" Creates a new identity transform.
		"""
		if rotate is True:
			rotate = 'zca'
		if isinstance(rotate, str) and rotate not in ('zca', 'pca'):
			raise ValueError('Invalid whitening transformation: {}'
				.format(rotate))

		if rotate == 'zca' and not scale:
			warnings.warn('"zca" without scaling is an identity transform. '
				'Only "center" will have any effect.', UserWarning)

		self.center = center
		self.scale = scale
		self.rotate = rotate
		self.state = None
		self.transform = None

	###########################################################################
	def apply(self, data):
		""" Applies the normalization to a single sample.
		"""
		if self.transform is None:
			logger.warning('Normalization transform has not been loaded. '
				'Using an identity transform instead. Be sure to call '
				'`Normalize.learn()` before trying to use `apply()`!')
			return data
		return self.transform(data)

	###########################################################################
	def has_learned(self):
		""" Returns True if we have a normalization transform available.
		"""
		return self.state is not None

	###########################################################################
	def get_dimensionality(self):
		""" Returns the dimensionality of the normalization.
		"""
		return self.state['mean'].shape[0]

	###########################################################################
	def learn(self, data):
		""" Learns the normalization with the current data stream.

			# Arguments

			data: list of data vectors (or matrices)
		"""

		if len(data) > Normalize.APPROXIMATION_CUTOFF:
			logger.warning('Data set is large. Normalization will be '
				'approximate.')
			indices = numpy.random.permutation(
				len(data)
			)[:Normalize.APPROXIMATION_CUTOFF]
			data = data[indices]

		data = numpy.vstack(data)
		mean = data.mean(axis=0)

		stddev = data.std(axis=0, ddof=1)
		stddev = numpy.where(stddev, stddev, 1.)

		centered = data - mean

		# pylint: disable=invalid-name

		# Figure out if we need the full matrices
		M, N = centered.shape
		_, S, Vt = numpy.linalg.svd(centered, full_matrices=bool(M < N))

		s = numpy.array([1/s if s > S[0] * 1e-6 else 0. for s in S])
		s *= numpy.sqrt(data.shape[0] - 1)
		S = numpy.diag(s)

		# pylint: enable=invalid-name

		self.state = {
			'mean' : mean,
			'stddev' : stddev,
			'S' : S,
			'Vt' : Vt
		}

		self.transform = self._build_transform()

	###########################################################################
	def _build_transform(self):
		""" Creates the normalization transform function.
		"""
		if self.rotate:
			current_transform = self.state['Vt']

			# This is not mathematically necessary, but PCA involves rotation
			# around eigenvectors, which are only unique up to an overall sign.
			# So to make things repeatable
			if self.rotate == 'pca':
				reflect = numpy.array(
					[-1 if r[0] < 0 else 1 for r in current_transform]
				)
				current_transform = reflect * current_transform

			if self.scale:
				current_transform = self.state['S'].dot(current_transform)

			if self.rotate == 'zca':
				current_transform = (self.state['Vt'].T).dot(current_transform)

			if self.center:
				return lambda x: (x - self.state['mean']).dot(
					current_transform.T)
			else:
				return lambda x: ((x - self.state['mean']).dot(
					current_transform.T) + self.state['mean'])

		if self.scale:
			if self.center:
				return lambda x: \
					(x - self.state['mean']) / self.state['stddev']
			else:
				return lambda x: ( \
					(x - self.state['mean']) / self.state['stddev'] + \
					self.state['mean']
				)

		if self.center:
			return lambda x: x - self.state['mean']
		else:
			return lambda x: x

	###########################################################################
	def save(self, filename):
		""" Saves normalization statistics.
		"""
		logger.info('Saving normalization state to: %s', filename)
		if self.state is None:
			to_save = {}
		else:
			to_save = {
				key : value.tolist() if hasattr(value, 'tolist') else value
				for key, value in self.state.items()
			}

		with open(filename, 'w') as fh:
			fh.write(yaml.dump(to_save))

	###########################################################################
	def restore(self, filename):
		""" Loads previously saved normalization statistics.
		"""
		logger.info('Restoring normalization state from: %s', filename)
		with open(filename, 'r') as fh:
			to_load = yaml.load(fh.read())

		if not to_load:
			self.state = None
			self.transform = None
		else:
			self.state = {
				key : numpy.array(value) if isinstance(value, (tuple, list)) \
					else value
				for key, value in to_load.items()
			}
			self.transform = self._build_transform()

	###########################################################################
	def get_state(self):
		""" Getter for the state.
		"""
		return self.state

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class BatchNormalization(Layer):	# pylint: disable=too-few-public-methods
	""" A batch normalization layer, which normalizes activations.
	"""

	###########################################################################
	@classmethod
	def get_container_name(cls):
		""" Returns the name of the container class.
		"""
		return 'batch_normalization'

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new batch normalization layer.
		"""
		super().__init__(*args, **kwargs)
		self.axis = None

	###########################################################################
	def _parse(self, engine):
		""" Parse the layer.
		"""
		if isinstance(self.args, dict):
			if 'axis' in self.args:
				self.axis = engine.evaluate(self.args['axis'], recursive=True)
				if not isinstance(self.axis, int):
					raise ParsingError('"axis" must be an integer.')

	###########################################################################
	def _build(self, model):
		""" Instantiates the layer with the given backend.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			if backend.keras_version() == 1:
				import keras.layers as L		# pylint: disable=import-error
				yield L.BatchNormalization(
					mode=2,
					axis=-1 if self.axis is None else self.axis,
					name=self.name
				)
			else:
				import keras.layers.normalization as L # pylint: disable=import-error
				yield L.BatchNormalization(
					axis=-1 if self.axis is None else self.axis,
					center=True,
					scale=True,
					name=self.name
				)

		elif backend.get_name() == 'pytorch':

			import torch.nn as nn				# pylint: disable=import-error
			from kur.backend.pytorch.modules import swap_channels

			def connect(inputs):
				""" Connects the layer
				"""
				assert len(inputs) == 1

				ndim = len(inputs[0]['shape'])
				func = {
					0 : nn.BatchNorm1d,
					1 : nn.BatchNorm1d,
					2 : nn.BatchNorm2d,
					3 : nn.BatchNorm3d
				}.get(ndim-1)
				if func is None:
					raise ValueError('Unsupported dimensionality: {}'
						.format(ndim))

				output = inputs[0]['layer']

				if ndim != 1:
					output = model.data.add_operation(
						swap_channels
					)(output)

				output = model.data.add_layer(
					self.name,
					func(inputs[0]['shape'][-1], affine=True)
				)(output)

				if ndim != 1:
					output = model.data.add_operation(
						swap_channels
					)(output)

				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : output
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Batch normalizations only take a single input.')
		input_shape = input_shapes[0]
		return input_shape

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys
import io
import re
import logging
from collections import OrderedDict

from . import Backend
from ..utils import can_import, idx, DisableLogging
from ..loss import Loss
from ..providers import BatchProvider

logger = logging.getLogger(__name__)

###############################################################################
class PyTorchBackend(Backend):
	""" Backend which uses PyTorch.
	"""

	###########################################################################
	@classmethod
	def is_supported(cls):
		""" Returns True if this backend can be used.
		"""
		return can_import('torch')

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the backend class.

			This is used by the base class's factory method.
		"""
		return 'pytorch'

	###########################################################################
	def save(self, model, filename):
		""" Saves the model weights to the given filename.

			# Arguments

			model: Model instance. The model whose weights should be saved.
			filename: str. The filename to write the weights to.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.

			# Return value

			None
		"""
		path = os.path.expanduser(os.path.expandvars(filename))
		if os.path.exists(path):
			if not os.path.isdir(path):
				raise ValueError('Target weight exists, but it is not a '
					'directory. Kur expected a directory that it can work '
					'with. Please move or delete the existing path: {}'
					.format(path))

			for dirpath, _, filenames in os.walk(path):
				for this_file in filenames:
					if this_file.endswith('.kur'):
						os.unlink(os.path.join(dirpath, this_file))
		else:
			os.makedirs(path, exist_ok=True)

		for k, v in model.data.model.state_dict().items():
			layer_name, name = k.split('.', 1)
			target = os.path.join(
				path,
				'{}+{}.kur'.format(layer_name, name)
			)

			idx.save(target, v.cpu().numpy())

	###########################################################################
	def restore(self, model, filename):
		""" Load the model weights from the given filename.

			# Arguments

			model: Model instance. The model whose weights should be restored.
			filename: str. The filename to read the weights from.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.

			# Return value

			None
		"""

		# pylint: disable=import-error
		import torch
		# pylint: enable=import-error

		path = os.path.expanduser(os.path.expandvars(filename))
		if os.path.exists(path):
			if not os.path.isdir(path):
				raise ValueError('Target weight exists, but it is not a '
					'directory. Kur expected a directory that it can work '
					'with. Please move or delete the existing path: {}'
					.format(path))
		else:
			raise ValueError('Target weight directory does not exist: {}'
				.format(path))

		# Enumerate all of the saved tensors, organized like this:
		# tensors = {
		#	'layer_1_name' : {
		#		'weight_1_name'  : '/path/to/file',
		#		...,
		#	},
		#	...
		# }
		tensors = self.enumerate_saved_tensors(path)

		state = dict(model.data.model.state_dict())

		for k in state:
			layer_name, name = k.split('.', 1)
			layers = tensors.get(layer_name)
			if not layers:
				logger.warning('A model layer was not restored because there '
					'was no corresponding weight file: %s', k)
				continue
			target = layers.get(name)
			if target is None:
				logger.warning('Part of a model layer was not restored '
					'because there as no corresponding weight file: %s', k)
				continue

			state[k] = torch.from_numpy(idx.load(target))

			del layers[name]

		remaining = [v for k, V in tensors.items() for v in V]
		if remaining:
			logger.warning('Some weight files were not used when restoring '
				'the model: %s', remaining)

		model.data.model.load_state_dict(state)

	###########################################################################
	def enumerate_saved_tensors(self, path):
		""" Enumerates saved tensors (weights).
		"""
		result = {}

		regex = re.compile(r'^(?P<layer>.*?)\+(?P<weight>.*?)\.kur$')
		for dirpath, dirnames, filenames in os.walk(path): # pylint: disable=unused-variable
			for filename in filenames:
				match = regex.match(filename)
				if match is None:
					continue
				filename = os.path.join(dirpath, filename)
				layer, weight = match.groups()
				if layer not in result:
					result[layer] = {}
				if weight in result[layer]:
					logger.warning('Tensor weights have already been loaded '
						'for layer="%s", tensor="%s" from file: %s. We will '
						'skip this new file we just found: %s.', layer, weight,
						result[layer][weight], filename)
					continue
				result[layer][weight] = filename

		return result

	###########################################################################
	def create_data(self, model):
		""" Requests a new set of model-specific data to be used during model
			assembly.
		"""
		from .pytorch.modules import TorchModel
		data = TorchModel(gpu=self.devices)
		data.allow_reuse = True
		return data

	###########################################################################
	def connect(self, inputs, target, data):
		""" Applies a tensor operation to a set of input tensors.

			# Arguments

			inputs: list. A list of input tensors.
			target: object. The tensor operation to apply.

			# Notes

			The exact underlying types of the `inputs` and `target` tensors
			depend on the backend implementation.

			# Return value

			A new tensor object (backend-specific) resulting from the applied
			operation.
		"""
		if not isinstance(inputs, list):
			inputs = [inputs]
		return target(inputs)

	###########################################################################
	def process_loss(self, model, loss):
		""" Process the loss functions.

			# Arguments

			model: The Kur model. It must be compiled.
			loss: Loss instance, list/tuple of Loss instances, or a dictionary
				of model layer names mapped to Loss instances.
		"""
		if loss is None:
			num_outputs = len(model.outputs)
			logger.error('You are trying to construct a training/validation'
				'/testing model, but you haven\'t specified any loss '
				'functions. Your model has %d outputs: %s. You need to '
				'specify %d loss functions, one for each output.',
				num_outputs, ', '.join(model.outputs), num_outputs)
			raise ValueError('No loss functions were specified, but are '
				'required for training, testing, and validation.')

		if isinstance(loss, Loss):
			loss = [loss]

		if len(loss) != len(model.outputs):
			raise ValueError('Model has {} outputs, but only {} loss '
				'functions were specified.'
				.format(len(model.outputs), len(loss)))

		if isinstance(loss, (list, tuple)):
			loss = dict(zip(model.outputs, loss))

		if not isinstance(loss, (dict, OrderedDict)):
			raise ValueError('Loss functions given to "compile" should be '
				'a list/tuple, a dictionary, or a single Loss instance. '
				'Instead we received this: {} (type={})'
				.format(loss, type(loss)))

		return OrderedDict(
			(
				target,
				this_loss.get_loss(model, target, model.outputs[target].value)
			)
			for target, this_loss in loss.items()
		)

	###########################################################################
	def compile(self, model, loss=None, optimizer=None, blocking=True,
		assemble_only=False):
		""" Creates an implementation-specific representation of the
			instantiated model.

			# Arguments

			model: Model instance. The built model to create a representation
				for.
			loss: Loss instance, list, dict, or None. The loss function
				(objective function) to use for training. If None, no loss
				functions are used. If a list, it should be a list of Loss
				instances, one for each model output; if the list is length 1,
				then the same loss function is used for each model output. If
				this is a dict, it should map from model output names to Loss
				instances to apply to that output.
			optimizer: Optimizer instance or None. The loss function to use for
				training. If None, no optimizer is used.
			blocking: bool (default: True). If True, requests that the
				implementation not return until all code compiling is complete.
				If False, the implementation may spawn background threads or
				processes, and these may still be running when
				train/test/evaluate are called (leading to blocking behavior at
				that point).
			assemble_only: bool (default: False). If True, does not attempt to
				create any compiled functions/models at all. Simply assembles
				the model and prepares the data sources. If this is True, then
				blocking is ignored.

			# Return value

			An implementation-specific object ("model") that can be used for
			training or evaluation. However, if `assemble_only` is True, then
			this returns a limited object which cannot be actually used for
			anything.

			# Notes

			Implementations may need to modify the Model. If this is the case,
			ExtensionState should be used to protect the model from extensions
			that are left stuck on after compiling.
		"""
		if model.compiled is None:
			model.compiled = {}

		model.data.set_outputs(
			[(k, v.value['layer']) for k, v in model.outputs.items()]
		)

		if loss is None and optimizer is None:
			key = 'evaluate'
			result = {}
		elif optimizer is None:
			torch_losses = self.process_loss(model, loss)

			key = 'test'
			result = {'loss' : torch_losses}
		else:
			torch_optimizer = optimizer.get_optimizer(self)(
				model.data.model.parameters())
			torch_losses = self.process_loss(model, loss)

			key = 'train'
			result = {'loss' : torch_losses, 'optimizer' : torch_optimizer}

		result.update({
			'model' : model.data
		})

		if logger.isEnabledFor(logging.DEBUG):
			x = io.StringIO()
			self.summary(model, x)
			for line in x.getvalue().split('\n'):
				logger.debug(line)

		if not assemble_only:
			model.compiled[key] = result
			if blocking:
				self.wait_for_compile(model, key)

		return result

	###########################################################################
	def wait_for_compile(self, model, key):
		""" Waits for the PyTorch model to be ready to roll.

			# Notes:

			PyTorch models are not compiled in the same way Theano/TensorFlow
			models are prepared. However, if you are using the GPU, then there
			is still overhead associated with spinning up CUDA. Theano and
			TensorFlow do this shortly after being initialized, but PyTorch
			waits until the last necessary minute. So although this call might
			more accurately be called `wait_for_cuda_devices`, we keep the name
			`wait_for_compile` to keep the API similar to, e.g., the Keras
			backend.
		"""
		if model.provider is None:
			logger.warning('No data provider available, so we cannot reliably '
				'wait for compiling to finish.')
			return

		with DisableLogging():
			provider = BatchProvider(
				sources=dict(zip(model.provider.keys, model.provider.sources)),
				batch_size=2*min(1, self.parallel),
				num_batches=1,
				randomize=False
			)
		model.supplement_provider(provider)

		logger.info('Waiting for model to be ready to use...')
		for batch in provider:
			model.data.predict(batch)

	###########################################################################
	def summary(self, model, file=None):
		""" Prints a model summary
		"""
		file = file or sys.stdout
		fill = lambda x, w: x[:min(w, len(x))].ljust(w)
		num_param = 0
		print('{}-+-{}-+-{}'.format('-'*30, '-'*20, '-'*10), file=file)
		print('{} | {} | {}'.format(
			fill('Layer Name', 30),
			fill('Shape', 20),
			fill('Parameters', 10)
		), file=file)
		print('{}-+-{}-+-{}'.format('-'*30, '-'*20, '-'*10), file=file)
		for k, v in model.data.model.state_dict().items():
			print('{} | {} | {}'.format(
				fill(k, 30),
				fill(str(tuple(v.size())), 20),
				fill(str(v.numel()), 10)
			), file=file)
			print('{}-+-{}-+-{}'.format('-'*30, '-'*20, '-'*10), file=file)
			num_param += v.numel()
		print('Total parameters: {}'.format(num_param), file=file)

	###########################################################################
	def train(self, model, data):
		""" Fits the given model on a batch of data.

			# Arguments

			model: Model instance. The model to train with.
			data: dict or list. The data to train a batch on.
			compiled: object. The data package returned by `compile`. Its type
				and content are backend-specific.

			# Return value

			A dictionary of loss values, whose keys are the names of the
			respective outputs. If per-output loss is not available, then a
			single, global loss value can be returned instead.
		"""

		torch_model = model.compiled['train']['model']
		optimizer = model.compiled['train']['optimizer']
		losses = model.compiled['train']['loss']

		optimizer.zero_grad()

		predictions, losses = torch_model.test(data, losses)

		torch_model.backprop(losses)

		optimizer.step()

		metrics = {
			k : loss.data.cpu().numpy().squeeze(-1)
			for k, loss in zip(model.outputs, losses)
		}

		predictions = {
			k : v.data.cpu().numpy() for k, v in zip(model.outputs, predictions)
		}

		return (predictions, metrics)

	###########################################################################
	def test(self, model, data):
		""" Calculates the model loss on a batch of data.

			# Arguments

			model: Model instance. The model to evaluate with.
			data: dict or list. The batch of data to test on.
			compiled: object. The data package returned by `compile`. Its type
				and content are backend-specific.

			# Return value

			A dictionary of loss values, whose keys are the names of the
			respective outputs. If per-output loss is not available, then a
			single, global loss value can be returned instead.
		"""
		torch_model = model.compiled['test']['model']
		losses = model.compiled['test']['loss']

		predictions, losses = torch_model.test(data, losses)

		metrics = {
			k : loss.data.cpu().numpy().squeeze(-1)
			for k, loss in zip(model.outputs, losses)
		}

		predictions = {
			k : v.data.cpu().numpy() for k, v in zip(model.outputs, predictions)
		}

		return (predictions, metrics)

	###########################################################################
	def evaluate(self, model, data):
		""" Evaluates the model on a batch ofdata.
		"""
		torch_model = model.compiled['evaluate']['model']

		predictions = torch_model.predict(data)

		predictions = {
			k : v.data.cpu().numpy() for k, v in zip(model.outputs, predictions)
		}

		return (predictions, {})

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from unittest.mock import patch

import pytest

from kur.kurfile import Kurfile
from kur.loss import Ctc
from kur.model import Executor

from modkurfile import modify_kurfile
from example import example

Executor.NAN_IS_FATAL = False
Executor.DEFAULT_RETRY_ENABLED = False

def replacement_ctc(self, model, target, output):
	""" Patch method on the CTC loss function.
	"""
	if model.get_backend().get_name() == 'pytorch':
		pytest.xfail('Backend "pytorch" does not use a CTC loss function.')
	return replacement_ctc.original(self, model, target, output)
replacement_ctc.original = Ctc.get_loss

@pytest.fixture
def example_directory():
	return os.path.normpath(
		os.path.join(
			os.path.dirname(__file__),
			'../..',
			'examples'
		)
	)

@pytest.fixture(
	params=['mnist.yml', 'cifar.yml', 'speech.yml']
)
def kurfile(request, example_directory, jinja_engine):
	result = Kurfile(
		os.path.join(example_directory, request.param),
		jinja_engine
	)

	modify_kurfile(result.data)
	for k in ('train', 'validate', 'test', 'evaluate'):
		if k in result.data and 'data' in result.data[k]:
			for data_source in result.data[k]['data']:
				if 'speech_recognition' in data_source \
					and 'normalization' in data_source['speech_recognition']:
					del data_source['speech_recognition']['normalization']

	result.parse()
	return result

@patch('kur.loss.Ctc.get_loss', replacement_ctc)
class TestExample:

	@example
	def test_train(self, kurfile):
		if 'train' not in kurfile.data:
			pytest.skip('No training section defined for this Kurfile.')
		func = kurfile.get_training_function()
		func()

	@example
	def test_test(self, kurfile):
		if 'test' not in kurfile.data:
			pytest.skip('No testing section defined for this Kurfile.')
		func = kurfile.get_testing_function()
		func()

	@example
	def test_evaluate(self, kurfile):
		if 'evaluate' not in kurfile.data:
			pytest.skip('No evaluation section defined for this Kurfile.')
		func = kurfile.get_evaluation_function()
		func()

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from unittest.mock import Mock
from functools import partial

import pytest

import numpy

from kur.supplier.speechrec import SpeechRecognitionSupplier

@pytest.fixture
def seed():
	return 0

@pytest.fixture
def num_entries():
	return 50

@pytest.fixture
def fake_data(num_entries):
	return list(range(num_entries))

@pytest.fixture
def fake_supplier(seed, fake_data):
	result = Mock()
	result.metadata = {'entries' : len(fake_data)}
	result.data = {'data' : fake_data}
	result.kurfile.get_seed.return_value = seed
	result.downselect = partial(SpeechRecognitionSupplier.downselect, result)
	return result

@pytest.fixture
def permuted_numbers(seed, num_entries):
	return numpy.random.RandomState(seed).permutation(num_entries)

def select_from(data, indices):
	if isinstance(data, numpy.ndarray):
		return data[sorted(indices)]

	indices = set(indices)
	return type(data)(x for i, x in enumerate(data) if i in indices)

class TestSpeechRecognitionSupplier:

	def verify(self, supplier, expected_length, expected_data):
		assert supplier.metadata['entries'] == expected_length
		assert numpy.allclose(supplier.data['data'], expected_data)

	def test_sample_none(self, fake_supplier, fake_data):
		fake_supplier.downselect(None)
		self.verify(fake_supplier, len(fake_data), fake_data)

	def test_sample_int(self, fake_supplier, fake_data, permuted_numbers):
		fake_supplier.downselect(10)
		self.verify(fake_supplier, 10,
			select_from(fake_data, permuted_numbers[:10]))

	def test_sample_percent(self, fake_supplier, fake_data, permuted_numbers):
		fake_supplier.downselect('10%')
		self.verify(fake_supplier, 5,
			select_from(fake_data, permuted_numbers[:5]))

	def test_sample_no_end(self, fake_supplier, fake_data, permuted_numbers):
		fake_supplier.downselect('10-')
		self.verify(fake_supplier, 40,
			select_from(fake_data, permuted_numbers[10:]))

	def test_sample_range(self, fake_supplier, fake_data, permuted_numbers):
		fake_supplier.downselect('10-20')
		self.verify(fake_supplier, 10,
			select_from(fake_data, permuted_numbers[10:20]))

	def test_sample_range_percent(self, fake_supplier, fake_data,
			permuted_numbers):
		fake_supplier.downselect('10-20%')
		self.verify(fake_supplier, 5,
			select_from(fake_data, permuted_numbers[5:10]))

	def test_sample_range_percent_no_end(self, fake_supplier, fake_data,
			permuted_numbers):
		fake_supplier.downselect('10-%')
		self.verify(fake_supplier, 45,
			select_from(fake_data, permuted_numbers[5:]))

	def test_sample_lots(self, fake_supplier, fake_data,
			permuted_numbers):
		fake_supplier.downselect('10000')
		self.verify(fake_supplier, 50,
			select_from(fake_data, permuted_numbers))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Repeat(Layer):				# pylint: disable=too-few-public-methods
	""" A layer which repeats its input a fixed number of times.
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new repeat layer.
		"""
		super().__init__(*args, **kwargs)
		self.count = None

	###########################################################################
	def _parse(self, engine):
		""" Parse the layer.
		"""
		if isinstance(self.args, dict):
			self.count = engine.evaluate(self.args['count'], recursive=True)
		else:
			self.count = self.args

		try:
			self.count = int(self.count)
		except ValueError:
			raise ParsingError('Key "count" in Repeat layer must be an '
				'integer. Received: {}'.format(self.count))

		if self.count < 1:
			raise ParsingError('Key "count" in Repeat layer must be >= 1.')

	###########################################################################
	def _build(self, model):
		""" Create the backend-specific placeholder.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			import keras.layers as L			# pylint: disable=import-error

			yield L.RepeatVector(
				self.count,
				name=self.name
			)

		elif backend.get_name() == 'pytorch':

			def connect(inputs):
				""" Connects the layer.
				"""
				assert len(inputs) == 1

				ndim = len(inputs[0]['shape'])
				sizes = (1, self.count) + (1, )*ndim
				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : model.data.add_operation(
						lambda x: x.unsqueeze(1).repeat(*sizes)
					)(inputs[0]['layer'])
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Repeat layers only take a single input.')
		input_shape = input_shapes[0]
		if len(input_shape) != 1:
			raise ValueError('Repeat layers only accept flat (1D) inputs.')
		return (self.count, input_shape[0])

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .hooks import TrainingHook, EvaluationHook
from .output_hook import OutputHook
from .mnist_hook import MnistHook
from .transcript import TranscriptHook
from .slack_hook import SlackHook
from .plot_hook import PlotHook
from .text_hook import TextHook
from .kurhub_hook import KurhubHook

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF
